#!/usr/bin/env python3

# Copyright Toolkit Authors (Yusuke Adachi)

import argparse
import logging
import os
import pickle
import shutil


def change_path_df(in_pkl, out_pkl):
    """Change absolute path to relative path.

    Args:
        in_pkl (str): Input picke file.
        out_pkl (str): Output pickle file.

    """
    in_pkl_path = os.path.abspath(in_pkl)
    base_dir_path = os.path.dirname(in_pkl_path)
    logging.info("Loading %s" % in_pkl)
    with open(in_pkl, 'rb') as f:
        df_dict = pickle.load(f)

    for key in df_dict:
        try:
            df_dict[key]['path'] = df_dict[key]['path'].str.replace(base_dir_path, '.')
        except KeyError:
            logging.info("Skipped. %s has no path info." % key)

    if os.path.isfile(out_pkl):
        bak_pkl = out_pkl + '.bak'
        logging.info("Backup %s to %s" % (out_pkl, bak_pkl))
        shutil.move(out_pkl, bak_pkl)
    with open(out_pkl, "wb") as f:
        pickle.dump(df_dict, f, protocol=2)
    logging.info("Save to %s" % out_pkl)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="This script change absolute path \
                                                  to relative path.")
    parser.add_argument('--in-pkl', help='Pickle files to be changed.')
    parser.add_argument('--out-pkl', help='Output pickle file.')
    parser.add_argument('-v', '--verbose', action='store_true', help='Verbose mode.')
    args = parser.parse_args()
    pkl_list, out_pkl = args.in_pkl, args.out_pkl

    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    change_path_df(pkl_list, out_pkl)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright Toolkit Authors

"""Search engine for time-series data."""

from .. import TimeSeriesCassandraDBHandler as _TimeSeriesCassandraDBHandler
from .. import StatisticsCassandraDBHandler as _StatisticsCassandraDBHandler
from . import BaseDBSearchEngine as _BaseDBSearchEngine
from . import register_engine


@register_engine(db_handlers=[
    _TimeSeriesCassandraDBHandler,
    _StatisticsCassandraDBHandler
])
class TimeSeriesCassandraDBSearchEngine(_BaseDBSearchEngine):
    """Search engine for V3TimeSeriesCassandraDB."""

    def __init__(self, db_handler: _TimeSeriesCassandraDBHandler):
        """Initialize TimeSeriesCassandraDBSearchEngine.

        Args:
            db_handler (V3TimeSeriesCassandraDBHandler): database handler

        """
        super(TimeSeriesCassandraDBSearchEngine, self).__init__(db_handler)
        self._span = db_handler._span

    def add_condition(self, condition: str):
        """Add a search condition.

        Args:
            condition (str): search condition
                             (e.g. "'/vehicle/acceleration/accel_linear_x/mean' > 0")

        """
        if 'record_id' in condition:
            if '=' not in condition:
                raise ValueError('record_id must be filtered using operator "=" '
                                 "(e.g. record_id = '242_16000000080000000570' )")
            if 'like' in condition.lower():
                raise ValueError('record_id must be filtered using operator "=" '
                                 "(e.g. record_id = '242_16000000080000000570' )")
            if "'" not in condition:
                raise ValueError("string must be single-quoted "
                                 "(e.g. record_id = '242_16000000080000000570' )")
        else:
            if '"' not in condition:
                raise ValueError('column name must be double-quoted '
                                 '(e.g. "/vehicle/acceleration/accel_linear_x/mean" > 0)')
        self._conditions.append(condition)

    @property
    def query(self):
        """Return query.

        Returns:
            (str): query

        """
        if self.condition == '':
            return self.select
        return '{0} where {1} allow filtering'.format(self.select, self.condition)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright Toolkit Authors

from copy import deepcopy
import logging
import os
from pathlib import Path
from tqdm import tqdm

from . import BaseDBHandler as _BaseDBHandler


class MetaDBHandler(_BaseDBHandler):
    """Handler for metadb."""

    df_name = 'meta_df'

    def __init__(self, *args, **kwargs):
        super(MetaDBHandler, self).__init__(*args, **kwargs)
        self.base_dir_path = os.path.dirname(os.path.realpath(self.path_to_db_file))

    def _preprocess_item(self, data_in):
        """Preprocess.

        Args:
            data_in (dict): a dict containing metadata

        Returns:
            (list): list of serialized dicts

        """
        assert 'contents' in data_in.keys()
        assert isinstance(data_in['contents'], dict)

        # Convert absolute path to relative path
        if 'path' in data_in.keys():
            # noinspection PyTypeChecker
            data_path = Path(data_in['path'])
            try:
                relative_path = data_path.relative_to(self.base_dir_path)
                data_in['path'] = relative_path
            except ValueError as e:
                logging.warning('Could not resolve relative path to file: {}'.format(data_path))
                logging.warning(str(e))

        # Parse contents
        data_out = []
        for content_name, content_info in data_in['contents'].items():
            data = deepcopy(data_in)
            data['content'] = content_name
            data.update(content_info)
            del data['contents']
            data_out.append(data)

        return data_out

    def add_list_of_data(self, data_in, **kwargs):
        """Add listed data to db.

        Args:
            data_in (list): a list of dicts containing data

        """
        self.logger.info('(Preprocess) Pre-processing metadata')
        data_flat = []
        for data_item in tqdm(data_in):
            data_flat += self._preprocess_item(data_item)
        super().add_list_of_data(data_flat, **kwargs)

    def __next__(self):
        """Return the next item."""
        data = super().__next__()

        # Convert relative path to absolute path
        if 'path' in data.keys():
            data['path'] = os.path.join(self.base_dir_path, data['path'])

        return data

    def get_content_df(self):
        """Return content_df."""
        return self.content_df

    def get_file_df(self):
        """Return file_df."""
        return self.file_df

    def get_record_id_df(self):
        """Return record_id_df."""
        return self.record_id_df

    @property
    def content_df(self):
        """Return content_df.

        Columns: record_id, path, content, msg_type, tag

        """
        df = self.df[['record_id', 'path', 'content', 'msg_type', 'tags']]
        df = df.to_pandas_df(strings=True, virtual=True)
        df = df.rename(columns={"tags": "tag"})     # FIXME: use virtual-column before converting to pandas df
        df['tag'] = df['tag'].apply(lambda x: list(set(x.split(':'))))
        df['path'] = df['path'].apply(lambda x: os.path.join(self.base_dir_path, x))
        return df

    @property
    def file_df(self):
        """Return file_df.

        Columns: path, record_id, type, content_type, start_timestamp, end_timestamp

        """
        # FIXME: Use VAEX's aggregation function instead of pandas'
        df = self.df[['path', 'record_id', 'type', 'content_type', 'start_timestamp', 'end_timestamp']]
        df = df.to_pandas_df(strings=True, virtual=True)
        df = df.groupby(['path'], as_index=False).agg({
            'record_id': 'first',
            'type': 'first',
            'content_type': 'first',
            'start_timestamp': 'min',
            'end_timestamp': 'max',
        })
        df['path'] = df['path'].apply(lambda x: os.path.join(self.base_dir_path, x))
        return df

    @property
    def record_id_df(self):
        """Return record_id_df.

        Columns: 'record_id', 'duration', 'start_timestamp', 'end_timestamp', 'tags'

        """
        df = self.df[['record_id', 'start_timestamp', 'end_timestamp', 'tags']]
        df['duration'] = df.end_timestamp - df.start_timestamp
        df = df.to_pandas_df(strings=True, virtual=True)

        # FIXME: Use VAEX's aggregation function instead of pandas'
        df = df.groupby(['record_id'], as_index=False).agg({
            'duration': 'max',
            'start_timestamp': 'min',
            'end_timestamp': 'max',
            'tags': ':'.join
        })
        df['tags'] = df['tags'].apply(lambda x: list(set(x.split(':'))))
        return df



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright Toolkit Authors

"""Downsample preprocessing function."""

import numpy as np

from .preprocess import BasePreprocess


class Downsample(BasePreprocess):
    """Downsample processing."""

    def __init__(self, target_frame_rate, mode='skipping'):
        """Downsampling initialization.

        Args:
            target_frame_rate (float): target frame rate in Hz
            mode (str): 'skipping' or 'averaging'

        """
        super(Downsample, self)
        assert mode in ['skipping', 'averaging']
        self.target_frame_rate = target_frame_rate
        self.mode = mode

    def processing(self, timestamps, values):
        """Downsample data into the target sampling rate.

        Args:
            timestamps (ndarray): timestamps [sec]
            values (ndarray): signal to downsample

        Returns:
            downsampled_timestamps (ndarray): timestamps [sec]
            downsampled_values (ndarray): signal to downsample

        """
        downsampled_timestamps = []
        downsampled_values = []
        buffer_timestamps = []
        buffer_values = []
        fps_previous_index = 0
        for timestamp, value in zip(timestamps, values):
            fps_current_index = timestamp // (1.0 / float(self.target_frame_rate))
            if fps_current_index == fps_previous_index:
                if self.mode == 'averaging':
                    buffer_timestamps.append(timestamp)
                    buffer_values.append(value)
                continue
            else:
                fps_previous_index = fps_current_index

            if self.mode == 'skipping':
                downsampled_timestamps.append(timestamp)
                downsampled_values.append(value)
            elif self.mode == 'averaging':
                if len(buffer_values) > 1:
                    downsampled_timestamps.append(buffer_timestamps[int(len(buffer_timestamps) // 2)])
                    downsampled_values.append(np.array(buffer_values).mean(axis=0))
                else:
                    downsampled_timestamps.append(timestamp)
                    downsampled_values.append(value)
                buffer_timestamps = []
                buffer_values = []
            else:
                continue

        downsampled_timestamps = np.array(downsampled_timestamps)
        downsampled_values = np.array(downsampled_values)

        return downsampled_timestamps, downsampled_values



#!/usr/bin/env python3

# Copyright Toolkit Authors (Yusuke Adachi)


import fire
from functools import partial
import logging
from multiprocessing import Pool
import time
from tqdm import tqdm

from pydtk.db import V3DBHandler as DBHandler
from pydtk.io import BaseFileReader
from pydtk.statistics import BaseStatisticCalculation


def _analysis(*args, **kwargs):
    try:
        main(*args, **kwargs)
    except Exception as e:
        print(e)


def batch_analysis(database_id,
                   span=60.0,
                   num_jobs=1,
                   meta_db_base_dir=None,
                   meta_db_engine=None,
                   meta_db_host=None,
                   meta_db_username=None,
                   meta_db_password=None,
                   meta_db_name=None,
                   output_db_engine=None,
                   output_db_host=None,
                   output_db_username=None,
                   output_db_password=None,
                   output_db_name=None,
                   verbose=False):
    """Make Statistics Dataframe Table with all contents.

    Args:
        database_id (str): ID of the target database (e.g. "Driving Behavior Database")
        span (float): Size of divided frame[sec]
        num_jobs (int): Number of jobs to work in parallel
        meta_db_base_dir (str): base directory of path
        meta_db_engine (str): Database engine of metadata
        meta_db_host (str): HOST of database of metadata
        meta_db_username (str): Username for the database of metadata
        meta_db_password (str): Password for the database of metadata
        meta_db_name (str): Database name of metadata
        output_db_engine (str): Database engine for storing statistics data
        output_db_host (str): HOST of database of statistics data
        output_db_username (str): Username for the database of statistics data
        output_db_password (str): Password for the database of statistics data
        output_db_name (str): Database name of statistics data
        verbose (bool): Verbose mode

    """
    # Prepare method
    analysis = partial(_analysis,
                       database_id,
                       span=span,
                       meta_db_base_dir=meta_db_base_dir,
                       meta_db_engine=meta_db_engine,
                       meta_db_host=meta_db_host,
                       meta_db_username=meta_db_username,
                       meta_db_password=meta_db_password,
                       meta_db_name=meta_db_name,
                       output_db_engine=output_db_engine,
                       output_db_host=output_db_host,
                       output_db_username=output_db_username,
                       output_db_password=output_db_password,
                       output_db_name=output_db_name,
                       verbose=verbose)

    # Load meta DB
    meta_db_handler = DBHandler(
        db_class='meta',
        db_engine=meta_db_engine,
        db_host=meta_db_host,
        db_name=meta_db_name,
        db_username=meta_db_username,
        db_password=meta_db_password,
        database_id=database_id,
        base_dir_path=meta_db_base_dir
    )

    # Get list of contents
    contents = meta_db_handler.df['contents'].to_list()

    # Process one-by-one
    with Pool(num_jobs) as pool:
        pool.map(analysis, contents)


def main(database_id,
         q_content,
         span=60.0,
         meta_db_base_dir=None,
         meta_db_engine=None,
         meta_db_host=None,
         meta_db_username=None,
         meta_db_password=None,
         meta_db_name=None,
         output_db_engine=None,
         output_db_host=None,
         output_db_username=None,
         output_db_password=None,
         output_db_name=None,
         verbose=False):
    """Make Statistics Dataframe Table.

    Args:
        database_id (str): ID of the target database (e.g. "Driving Behavior Database")
        q_content (str): Content name of query
        span (float): Size of divided frame[sec]
        meta_db_base_dir (str): base directory of path
        meta_db_engine (str): Database engine of metadata
        meta_db_host (str): HOST of database of metadata
        meta_db_username (str): Username for the database of metadata
        meta_db_password (str): Password for the database of metadata
        meta_db_name (str): Database name of metadata
        output_db_engine (str): Database engine for storing statistics data
        output_db_host (str): HOST of database of statistics data
        output_db_username (str): Username for the database of statistics data
        output_db_password (str): Password for the database of statistics data
        output_db_name (str): Database name of statistics data
        verbose (bool): Verbose mode

    """
    if verbose:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.ERROR)

    t_n, t_b = time.time(), time.time()

    # Load meta DB
    meta_db_handler = DBHandler(
        db_class='meta',
        db_engine=meta_db_engine,
        db_host=meta_db_host,
        db_name=meta_db_name,
        db_username=meta_db_username,
        db_password=meta_db_password,
        database_id=database_id,
        base_dir_path=meta_db_base_dir,
        read_on_init=False
    )
    logging.info("Loading content: {}".format(q_content))
    meta_db_handler.read(where='contents like "{}"'.format(q_content))
    reader = BaseFileReader()
    calculator = BaseStatisticCalculation(span, sync_timestamps=True)
    t_n, t_p = time.time(), t_n
    logging.info("Loaded index and filtered files.({0:.03f} secs)".format(t_n - t_p))

    # Initialize DB-Handler
    stat_db_handler = DBHandler(
        db_class='statistics',
        db_engine=output_db_engine,
        db_host=output_db_host,
        db_name=output_db_name,
        db_username=output_db_username,
        db_password=output_db_password,
        database_id=database_id,
        span=span,
        read_on_init=False
    )

    # Read data and write calculated data in DB
    def write_stat_to_db(item):
        # Load data from file and get statistical table
        timestamps, data, columns = reader.read(metadata=item)
        stat_df = calculator.statistic_tables(timestamps, data, columns)
        stat_df.insert(0, "record_id", item["record_id"])

        # Write to DB
        stat_db_handler.df = stat_df
        stat_db_handler.save()

    tqdm.pandas(desc="Load files, calculate and write")
    for sample in meta_db_handler:
        write_stat_to_db(sample)
    t_n, t_p = time.time(), t_n
    logging.info("Calculated statistics and wrote to DB.({0:.03f} secs)".format(t_n - t_p))

    logging.info("Done.(Total: {0:.03f} secs)".format(t_n - t_b))


def batch_script():
    """Function for tool.poetry.scripts."""
    fire.Fire(batch_analysis)


def script():
    """Function for tool.poetry.scripts."""
    fire.Fire(main)


if __name__ == "__main__":
    fire.Fire(main)
    # fire.Fire(batch_analysis)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright Toolkit Authors

"""DB Engines for V4DBHandler."""

from copy import deepcopy
import logging
from typing import Optional

from pymongo import MongoClient, UpdateOne, DeleteOne
from ..deps import pql as PQL

logger = logging.getLogger(__name__)

DEFAULT_DB_NAME = 'default'
DEFAULT_COLLECTION_NAME = 'default'


def connect(
    db_host: str,
    db_name: Optional[str] = DEFAULT_DB_NAME,
    db_username: Optional[str] = None,
    db_password: Optional[str] = None,
    collection_name: Optional[str] = DEFAULT_COLLECTION_NAME,
    **kwargs
):
    """Connect to DB.

    Args:
        db_host (str): database host
        db_name (str): database name
        db_username (str): database username
        db_password (str): database password
        collection_name (str): collection name

    Returns:
        (any): connection

    """
    if db_name is None:
        db_name = DEFAULT_DB_NAME
    if collection_name is None:
        collection_name = DEFAULT_COLLECTION_NAME

    if ':' in db_host:
        address = db_host.split(':')[0]
        port = int(db_host.split(':')[1])
    else:
        address = db_host
        port = None

    kwargs = {}
    if db_name is not None:
        kwargs.update({'authSource': 'admin'})
    connection = MongoClient(
        host=address,
        port=port,
        username=db_username,
        password=db_password,
        **kwargs
    )
    db = getattr(connection, db_name)
    collection = getattr(db, collection_name)
    return collection


def read(db,
         query: Optional[dict] = None,
         pql: any = None,
         group_by: Optional[str] = None,
         order_by: Optional[str] = None,
         limit: Optional[int] = None,
         offset: Optional[int] = None,
         handler: any = None,
         disable_count_total: bool = False,
         **kwargs):
    """Read data from DB.

    Args:
        db (Collection): DB connection
        query (dict or Query): Query to select items
        pql (PQL) Python-Query-Language to select items
        group_by (str): Aggregate by this key
        order_by (list): column name to sort by with format [ ( column1, 1 or -1 ), ... ]
        limit (int): number of items to return per a page
        offset (int): offset of cursor
        handler (BaseDBHandler): DBHandler
        disable_count_total (bool): set True to avoid counting total number of records
        **kwargs: kwargs for function `pandas.read_sql_query`
                  or `influxdb.DataFrameClient.query`

    Returns:
        (list, int): list of data and total number of records

    """
    if limit is None:
        limit = 0
    if offset is None:
        offset = 0

    if pql is not None and query is not None:
        raise ValueError('Either query or pql can be specified')

    if pql:
        query = PQL.find(pql)

    if group_by is None:
        if query:
            if order_by is None:
                data = db.find(query).skip(offset).limit(limit)
                count_total = db.count(query) if not disable_count_total else None
            else:
                data = db.find(query).sort(order_by).skip(offset).limit(limit)
                count_total = db.count(query) if not disable_count_total else None
        else:
            if order_by is None:
                data = db.find().skip(offset).limit(limit)
                count_total = db.count({}) if not disable_count_total else None
            else:
                data = db.find().sort(order_by).skip(offset).limit(limit)
                count_total = db.count({}) if not disable_count_total else None
    else:
        aggregate = []
        if query:
            aggregate.append({'$match': query})

        columns = {}
        for column in set(handler.columns).union(['_uuid', '_creation_time']):
            try:
                config = next(filter(lambda c: c['name'] == column, handler.config['columns']))
                agg = config['aggregation']
                columns.update({column: {'${}'.format(agg): '${}'.format(column)}})
            except Exception:
                columns.update({column: {'$first': '${}'.format(column)}})

        aggregate.append({
            '$group': {
                **columns,
                '_id': '${}'.format(group_by),
            }
        })
        aggregate.append({'$project': {'_id': 0}})
        aggregate_count = deepcopy(aggregate)
        aggregate_count.append({'$count': 'count'})

        if order_by is not None:
            aggregate.append({'$sort': {item[0]: item[1] for item in order_by}})

        if offset > 0:
            aggregate.append({'$skip': offset})
        if limit > 0:
            aggregate.append({'$limit': limit})

        data = db.aggregate(aggregate, allowDiskUse=True)
        try:
            count_total = list(db.aggregate(aggregate_count))[0]['count'] \
                if not disable_count_total else None
        except Exception as e:
            logger.warning(e)
            count_total = None

    data = list(data)
    count_total = count_total if count_total is not None else len(data)

    return data, count_total


def upsert(db, data, **kwargs):
    """Write data to DB.

    Args:
        db (Collection): DB connection
        data (list): data to save

    """
    ops = [
        UpdateOne({'_uuid': record['_uuid']}, {'$set': record}, upsert=True)
        for record in data
    ]
    if len(ops) > 0:
        db.bulk_write(ops)


def remove(db, uuids, **kwargs):
    """Remove data from DB.

    Args:
        db (Collection): DB connection
        uuids (list): A list of unique IDs

    """
    ops = [
        DeleteOne({'_uuid': uuid})
        for uuid in uuids
    ]
    if len(ops) > 0:
        db.bulk_write(ops)


def drop_table(db, name, **kwargs):
    """Drop a table from DB.

    Args:
        db (Collection): DB connection
        name (str): Name of the target table

    """
    db.database.__getitem__(name).drop()


def exist_table(db, name, **kwargs):
    """Check if the specified table (collection) exist.

    Args:
        db (Collection): DB connection
        name (str): Name of the target table (collection)

    """
    return name in db.database.list_collection_names()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright Toolkit Authors

from abc import ABC

from pydtk.models import BaseModel, register_model
import json
import numpy as np


@register_model(priority=1)
class GenericJsonModel(BaseModel, ABC):
    """A generic model for a json file."""

    _content_type = None    # allow any content-type
    _data_type = None  # allow any data-type
    _file_extensions = [".json"]
    _contents = None

    def __init__(self, **kwargs):
        super(GenericJsonModel, self).__init__(**kwargs)

    def _load(self, path, **kwargs):
        """Load a json file.

        Args:
            path (str): path to a json file
        """
        with open(path, "r") as fp:
            data = json.load(fp)
        self.data = data

    def _save(self, path, **kwargs):
        """Save dict data to a json file.

        Args:
            path (str): path to the output json file

        """
        with open(path, "w") as fp:
            json.dump(self.data, fp)

    @property
    def timestamps(self):
        """Return timestamps as ndarray."""
        if "time_stamps" in self.data.keys():
            return np.ndarray(self.data["time_stamps"])

        return np.ndarray([])

    def to_ndarray(self):
        """Json model do not support to_ndarray."""
        raise AttributeError(
            "JsonModel do not support to_ndarray, since it may be semi-structured"
        )

    @classmethod
    def generate_contents_meta(cls, path, content_key='content'):
        """Generate contents metadata.

        Args:
            path (str): File path
            content_key (str): Key of content

        Returns:
            (dict): contents metadata

        """
        # Load file
        with open(path, "r") as p:
            data = json.load(p)

        if 'contents' in data.keys():
            raise AttributeError(
                f"File '{path}' itself is a metadata"
            )

        # Generate metadata
        contents = {content_key: {'keys': list(data.keys()), 'tags': ['json']}}
        return contents

    @classmethod
    def generate_timestamp_meta(cls, path):
        """Generate contents metadata.

        Args:
            path (str): File path

        Returns:
            (list): [start_timestamp, end_timestamp]

        """
        raise NotImplementedError



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright Toolkit Authors

"""pydtk modules."""

from abc import ABCMeta, abstractmethod
import importlib
import json
import logging
import os
import re
import six
import pprint

from pydtk.utils.utils import dict_reg_match

MODELS_BY_PRIORITY = {}     # key: priority, value: model class

logger = logging.getLogger(__name__)


def register_models():
    """Register models."""
    for filename in os.listdir(os.path.join(os.path.dirname(__file__))):
        if filename == '__init__.py':
            continue

        try:
            importlib.import_module(os.path.join('pydtk.models',
                                    os.path.splitext(filename)[0]).replace(os.sep, '.'))
        except (ModuleNotFoundError, ImportError):
            logger.warning('Failed to load models in {}'.format(filename))


def register_model(priority=0):
    """Regist a model."""
    def decorator(cls):
        if priority not in MODELS_BY_PRIORITY.keys():
            MODELS_BY_PRIORITY.update({priority: []})
        MODELS_BY_PRIORITY[priority].append(cls)
        return cls

    return decorator


class UnsupportedFileError(BaseException):
    """Error for unsupported file."""

    pass


class MetaDataModel(object):
    """A model for a metadata file."""

    _file_extensions = ['.json']
    _data = dict()

    _key_map = {
        'content_type': 'content-type',
        'data_type': 'type',
    }
    _key_map_inv = {v: k for k, v in _key_map.items()}

    def __init__(self, data=None, **kwargs):
        super(MetaDataModel, self).__init__(**kwargs)
        if data is not None:
            if isinstance(data, MetaDataModel):
                self.data = data.data
            elif isinstance(data, dict):
                self.data = data
            else:
                raise TypeError('Unsupported type of data: {}'.format(type(data).__name__))

    @classmethod
    def is_loadable(cls, path='', **kwargs):
        """Check file format."""
        _, ext = os.path.splitext(path)
        if ext not in cls._file_extensions:
            return False
        return True

    def load(self, path, **kwargs):
        """Load json format metadata."""
        with open(path, 'r', encoding='utf-8') as f:
            data = json.load(f)
        self.data = data

    def save(self, path):
        """Save metadata as json."""
        _data = {}
        for key, value in self.data.items():
            if key not in self._key_map.keys():
                _data.update({key: value})
            else:
                _data.update({self._key_map[key]: value})

        with open(path, 'w') as f:
            json.dump(_data, f)

    @property
    def data(self):
        """Return data."""
        return self._data

    @data.setter
    def data(self, data):
        """Setter for self._data."""
        _data = {}
        for key, value in data.items():
            if key not in self._key_map_inv.keys():
                _data.update({key: value})
            else:
                _data.update({self._key_map_inv[key]: value})
        self._data = _data

    def to_dict(self):
        """Returns the model properties as a dict."""
        result = {}

        for attr, _ in six.iteritems(self.__dict__):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model."""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`."""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal."""
        if not isinstance(other, MetaDataModel):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal."""
        return not self == other


class BaseModel(metaclass=ABCMeta):
    """Base model.

    Priority for being selected as a reader of a file.
    The higher this value is, the more likely this model will be chosen as a file loader.
    """

    _priority = 0

    _content_type = None  # e.g. 'text/csv', 'text/.*'
    _data_type = None  # e.g. 'raw_data', '.*'
    _file_extensions = []  # e.g. ['mp4', 'avi']
    _contents = None  # Supported contents (e.g. {'camera/.*': {'tags': ['.*']}})
    _data = None    # for storing data
    _config = None  # for model configurations
    _metadata = None  # e.g. MetaDataModel()
    _columns = None  # Name of each columns in the ndarray returned by `to_ndarray`

    @abstractmethod
    def __init__(self, metadata=None, data=None, **kwargs):
        self.logger = logging.getLogger(name=type(self).__name__)
        if metadata is not None:
            self.metadata = metadata
        if data is not None:
            self.data = data

    def configure(self, **kwargs):
        """Configure this model.

        Args:
            **kwargs: settings (key: key in self._config, value: value of self._config[key])

        """
        for key, value in kwargs.items():
            if key in self._config.keys():
                self._config[key] = value
            else:
                raise KeyError('Unknown key: {}'.format(key))

    @classmethod
    def _is_loadable(cls, path='', contents=None, content_type=None, data_type=None, **kwargs):
        """Check data by file format."""
        return True

    @classmethod
    def is_loadable(cls, path='', contents=None, content_type=None, data_type=None, **kwargs):
        """Check if the given file is loadable.

        Args:
            path (str): path to the target file
            contents (dict or str): content to load (e.g. {'camera/front': 'tags': {...}})
            content_type (str): content-type (e.g. 'text/csv')
            data_type (str): data-type (e.g. 'raw_data')

        """
        # check by file extension
        _, ext = os.path.splitext(path)
        if ext not in cls._file_extensions:
            return False

        # check by content-type
        if cls._content_type is not None:
            if content_type is None:
                if not re.fullmatch(cls._content_type, ""):
                    return False
            else:
                if not re.fullmatch(cls._content_type, content_type):
                    return False

        # check by data-type
        if cls._data_type is not None:
            if data_type is None:
                if not re.fullmatch(cls._data_type, ""):
                    return False
            else:
                if not re.fullmatch(cls._data_type, data_type):
                    return False

        # check by contents
        if cls._contents is not None:
            if contents is None:
                if isinstance(cls._contents, str):
                    if re.fullmatch(cls._contents, "") is None:
                        return False
                if isinstance(cls._contents, list):
                    if any([re.fullmatch(_contents, "")
                            for _contents in cls._contents]) is False:
                        return False
                if isinstance(cls._contents, dict):
                    if re.fullmatch(next(iter(cls._contents)), "") is None:
                        return False
            else:
                if isinstance(contents, dict) and len(contents.keys()) > 1:
                    logging.warning('Loading multiple contents is not supported')
                    return False
                if isinstance(contents, list) and len(contents) > 1:
                    logging.warning('Loading multiple contents is not supported')
                    return False
                if isinstance(contents, list) and len(contents) == 1:
                    contents = contents[0]
                if isinstance(cls._contents, str) and isinstance(contents, str):
                    if re.fullmatch(cls._contents, contents) is None:
                        return False
                if isinstance(cls._contents, str) and isinstance(contents, dict):
                    if re.fullmatch(cls._contents, next(iter(contents))) is None:
                        return False
                if isinstance(cls._contents, list) and isinstance(contents, str):
                    if any([re.fullmatch(_contents, contents)
                            for _contents in cls._contents]) is False:
                        return False
                if isinstance(cls._contents, list) and isinstance(contents, dict):
                    if any([re.fullmatch(_contents, next(iter(contents)))
                            for _contents in cls._contents]) is False:
                        return False
                if isinstance(cls._contents, dict) and isinstance(contents, str):
                    if re.fullmatch(next(iter(cls._contents)), contents) is None:
                        return False
                if isinstance(cls._contents, dict) and isinstance(contents, dict):
                    if dict_reg_match(cls._contents, contents) is False:
                        return False

        # check data by file format
        if not cls._is_loadable(path=path, contents=contents, content_type=content_type,
                                data_type=data_type, **kwargs):
            return False

        # in case all check passed
        return True

    @property
    def data(self):
        """Gets the data of this class object.

        Returns:
            (object): any object

        """
        return self._data

    @data.setter
    def data(self, data):
        """Sets the data to this class object.

        Args:
            data (object): any object

        """
        self._data = data

    @property
    def metadata(self):
        """Gets the metadata of this class object.

        Returns:
            (object): a MetaDataModel object

        """
        return self._metadata

    @metadata.setter
    def metadata(self, metadata):
        """Sets the metadata to this class object.

        Args:
            metadata (any): dict or a MetaDataModel object

        """
        assert any([isinstance(metadata, dict), isinstance(metadata, MetaDataModel)])
        if isinstance(metadata, dict):
            metadata = MetaDataModel(data=metadata)
        self._metadata = metadata

    @abstractmethod
    def _load(self, path, contents=None, start_timestamp=None, end_timestamp=None, **kwargs):
        """Load data from a file.

        Args:
            path (str): path to the input file
            contents (dict or str): content to load
            start_timestamp (float): timestamp to start loading in sec.
            end_timestamp (float): timestamp to end loading in sec.

        """
        raise NotImplementedError

    def load(self, path=None, as_generator=False, **kwargs):
        """Load data from a file.

        Args:
            path (str): file-path to load (if None, one in the metadata will be used)
            content (str): content to load
            start_timestamp (float): timestamp to start loading in sec. (optional)
            end_timestamp (float): timestamp to end loading in sec. (optional)
            as_generator (bool): load data as a generator.

        """
        if path is None:
            if self.metadata is None:
                raise ValueError('filepath is not provided')
            else:
                path = self.metadata.data['path']

        # prepare metadata for loading
        metadata = self.metadata if self.metadata is not None else MetaDataModel()
        metadata.data.update({'path': path})

        # add extra information
        metadata.data.update(kwargs)

        # update metadata
        self.metadata = metadata

        # check if this model can load the data
        if not self.is_loadable(**metadata.data):
            raise UnsupportedFileError('Model "{0}" does not support loading file: {1}'.
                                       format(type(self).__name__, path))

        if as_generator:
            def load_as_generator():
                yield from self._load_as_generator(**metadata.data)
            return load_as_generator()
        else:
            self._load(**metadata.data)

    @abstractmethod
    def _save(self, path, **kwargs):
        """Save data to a file.

        Args:
            path (str): path to the output file

        """
        raise NotImplementedError

    def save(self, path=None, **kwargs):
        """Save data to a file."""
        if path is None:
            if self.metadata is None:
                raise ValueError('filepath is not provided')
            else:
                path = self.metadata.data['path']

        # check
        _, ext = os.path.splitext(path)
        if ext.lower() not in self._file_extensions:
            raise ValueError('File extension must be one of: {}'.format(self._file_extensions))

        # save
        self._save(path=path, **kwargs)

    @property
    def columns(self):
        """Return the name of each column in ndarray returned from `to_ndarray`."""
        postfixes = None
        content = 'unknown'

        # Get post-fixes
        if self._columns is not None:
            postfixes = self._columns

        # Get content
        try:
            contents = self.metadata.data['contents']
            if isinstance(contents, str):
                content = contents
            elif isinstance(contents, list):
                if len(contents) == 1:
                    content = contents[0]
                else:
                    raise ValueError('Loading multiple contents is not supported')
            elif isinstance(contents, dict):
                if len(contents.keys()) == 1:
                    content = list(contents.keys())[0]
                else:
                    raise ValueError('Loading multiple contents is not supported')
        except KeyError:
            raise

        # Create column names
        if postfixes is None:
            return [content]
        elif isinstance(postfixes, str):
            return [content + '/' + postfixes]
        elif isinstance(postfixes, list):
            return [content + '/' + postfix for postfix in postfixes]
        else:
            raise TypeError('Unable to handle variable `postfixes`.')

    def to_ndarray(self):
        """Return data as a ndarray.

        Returns:
            (ndarray): data

        """
        raise NotImplementedError

    @property
    def timestamps(self):
        """Return timestamps.

        Returns:
            (ndarray): timestamps

        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def generate_contents_meta(cls, path, content_key='content'):
        """Generate contents metadata.

        Args:
            path (str): File path
            content_key (str): Key of content

        Returns:
            (list): contents metadata

        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def generate_timestamp_meta(cls, path):
        """Generate contents metadata.

        Args:
            path (str): File path

        Returns:
            (list): [start_timestamp, end_timestamp]

        """
        raise NotImplementedError

    def to_dict(self):
        """Returns the model properties as a dict."""
        result = {}

        for attr, _ in six.iteritems(self.__dict__):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model."""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`."""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal."""
        if not isinstance(other, BaseModel):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal."""
        return not self == other

    def downsample_timestamps(self, timestamps, target_frame_rate=2.0):
        """Downsample timestamps into the target sampling rate.

        Args:
            timestamps (ndarray): timestamps [sec]

        Returns:
            downsampled_timestamps (ndarray): timestamps [sec]

        """
        downsampled_timestamps = timestamps
        fps_previous_index = 0
        for i, timestamp in enumerate(timestamps):
            fps_current_index = timestamp // (1.0 / float(target_frame_rate))
            if fps_current_index == fps_previous_index:
                downsampled_timestamps[i] = -1
                continue
            else:
                fps_previous_index = fps_current_index
        downsampled_timestamps = [timestamp for timestamp in downsampled_timestamps if timestamp != -1]

        return downsampled_timestamps


register_models()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright Toolkit Authors

"""DB Engines for V4DBHandler."""
from datetime import datetime
from typing import Optional

from montydb import MontyClient, set_storage
from ..deps import pql as PQL

DEFAULT_DB_NAME = 'default'
DEFAULT_COLLECTION_NAME = 'default'


def connect(
    db_host: str,
    db_name: Optional[str] = None,
    collection_name: Optional[str] = None,
    **kwargs
):
    """Connect to DB.

    Args:
        db_host (str): database host
        db_name (str): database name
        collection_name (str): collection name

    Returns:
        (MontyCollection): connection

    """
    if db_name is None:
        db_name = DEFAULT_DB_NAME
    if collection_name is None:
        collection_name = DEFAULT_COLLECTION_NAME

    set_storage(
        # general settings
        repository=db_host,  # dir path for database to live on disk, default is {cwd}
        storage='lightning',  # storage name, default "flatfile"
        use_bson=None,  # default None, and will import pymongo's bson if None or True

        # any other kwargs are storage engine settings.
        cache_modified=0,  # the only setting that flat-file have
    )

    collection = getattr(getattr(MontyClient(db_host), db_name), collection_name)
    return collection


def read(db,
         query: Optional[dict] = None,
         pql: any = None,
         order_by: Optional[str] = None,
         limit: Optional[int] = None,
         offset: Optional[int] = None,
         disable_count_total: bool = False,
         **kwargs):
    """Read data from DB.

    Args:
        db (MontyCollection): DB connection
        query (dict or Query): Query to select items
        pql (PQL) Python-Query-Language to select items
        order_by (list): column name to sort by with format [ ( column1, 1 or -1 ), ... ]
        limit (int): number of items to return per a page
        offset (int): offset of cursor
        disable_count_total (bool): set True to avoid counting total number of records
        **kwargs: kwargs for function `pandas.read_sql_query`
                  or `influxdb.DataFrameClient.query`

    Returns:
        (list, int): list of data and total number of records

    """
    if limit is None:
        limit = 0
    if offset is None:
        offset = 0

    if pql is not None and query is not None:
        raise ValueError('Either query or pql can be specified')

    if pql:
        query = PQL.find(pql)

    if query:
        query = _fix_query_exists(query)
        if order_by is None:
            data = db.find(query).skip(offset).limit(limit)
            count_total = db.count(query) if not disable_count_total else None
        else:
            data = db.find(query).sort(order_by).skip(offset).limit(limit)
            count_total = db.count(query) if not disable_count_total else None
    else:
        if order_by is None:
            data = db.find().skip(offset).limit(limit)
            count_total = db.count({}) if not disable_count_total else None
        else:
            data = db.find().sort(order_by).skip(offset).limit(limit)
            count_total = db.count({}) if not disable_count_total else None

    data = list(data)
    count_total = count_total if count_total is not None else len(data)

    return data, count_total


def upsert(db, data, **kwargs):
    """Write data to DB.

    Args:
        db (MontyCollection): DB connection
        data (list): data to save

    """
    for record in data:
        _record = _fix_datetime(record)
        uuid = _record['_uuid']
        existing_record = db.find_one({'_uuid': uuid})
        if existing_record is not None:
            existing_record.update(_record)
            db.replace_one({'_uuid': uuid}, existing_record)
        else:
            db.insert_one(_record)


def remove(db, uuids, **kwargs):
    """Remove data from DB.

    Args:
        db (TinyMongoCollection): DB connection
        uuids (list): A list of unique IDs

    """
    for uuid in uuids:
        db.delete_many({'_uuid': uuid})


def drop_table(db, name, **kwargs):
    """Drop a table from DB.

    Args:
        db (MontyCollection): DB connection
        name (str): Name of the target table

    """
    db.database.drop_collection(name)


def exist_table(db, name, **kwargs):
    """Check if the specified table (collection) exist.

    Args:
        db (MontyCollection): DB connection
        name (str): Name of the target table

    """
    return name in db.database.list_collection_names()


def _fix_query_exists(query):
    if isinstance(query, list):
        fixed_query = []
        for item in query:
            fixed_query.append(_fix_query_exists(item))
        return fixed_query

    elif isinstance(query, dict):
        fixed_query = {}
        for key, value in query.items():
            if isinstance(value, dict) or isinstance(value, list):
                fixed_query[key] = _fix_query_exists(value)
            elif key == '$exists' and value:
                fixed_query['$ne'] = None
            else:
                fixed_query[key] = value

        return fixed_query

    else:
        raise ValueError


def _fix_datetime(data: dict):
    """Fix datetime object.

    Args:
        data (dict): Input data

    Returns:
        (dict): Fixed data

    """
    _data = {}
    for key, value in data.items():
        if isinstance(value, datetime):
            _data[key] = value.timestamp()
        else:
            _data[key] = value

    return _data



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright Toolkit Authors

from typing import Optional

from copy import deepcopy
import hashlib
import logging
import os
from pathlib import Path

from flatten_dict import flatten

from . import BaseDBHandler as _BaseDBHandler
from . import register_handler


@register_handler(db_classes=['meta'], db_engines=['tinydb', 'tinymongo', 'mongodb', 'montydb'])
class MetaDBHandler(_BaseDBHandler):
    """Handler for metadb."""

    _database_id: str = ''
    _df_class = 'meta_df'

    def __init__(self,
                 database_id: Optional[str] = 'default',
                 base_dir_path: Optional[str] = None,
                 orient='path',
                 **kwargs):
        """Initialize MetaDBHandler.

        Args:
            database_id (str): ID of the database (e.g. "Driving Behavior Database")
            base_dir_path (str): base directory for path
            orient (str): key name over which to iterate

        """
        self._database_id = database_id
        self.orient = orient

        # Solve base-dir
        base_dir = base_dir_path
        if base_dir is None:
            base_dir = '/'
        self.base_dir_path = os.path.realpath(base_dir)

        # Prepare another DB-handler for storing a list of database_id
        self._database_id_db_handler = DatabaseIDDBHandler(**{**kwargs, 'read_on_init': False})

        # Prepare buffer for iteration
        self._buf = []
        self._indices = []  # [[record_idx, orient_idx]]
        self._indexed = False

        super(MetaDBHandler, self).__init__(**kwargs)

    def _initialize_engine(self,
                           db_engine: Optional[str] = None,
                           db_host: Optional[str] = None,
                           db_name: Optional[str] = None,
                           db_username: Optional[str] = None,
                           db_password: Optional[str] = None):
        """Initialize DB engine."""
        # Load settings from environment variables
        engine = db_engine if db_engine is not None \
            else os.environ.get('PYDTK_META_DB_ENGINE', None)
        username = db_username if db_username is not None \
            else os.environ.get('PYDTK_META_DB_USERNAME', None)
        password = db_password if db_password is not None \
            else os.environ.get('PYDTK_META_DB_PASSWORD', None)
        host = db_host if db_host is not None \
            else os.environ.get('PYDTK_META_DB_HOST', None)
        database = db_name if db_name is not None \
            else os.environ.get('PYDTK_META_DB_DATABASE', None)

        super()._initialize_engine(engine, host, database, username, password)

    def _solve_path(self, path: str, target: str):
        if target == 'relative':
            # noinspection PyTypeChecker
            data_path = Path(path)
            try:
                relative_path = data_path.relative_to(self.base_dir_path)
                path = str(relative_path)
            except ValueError as e:
                logging.warning('Could not resolve relative path to file: {}'.format(data_path))
                logging.warning(str(e))
        elif target == 'absolute':
            path = os.path.join(self.base_dir_path, path)
        else:
            raise ValueError('Unrecognized target: {}'.format(target))

        return path

    def _solve_path_in_data(self, data_in: dict, target: str):
        """Fix absolute path to relative one.

        Args:
            data_in (dict): a dict containing metadata
            target (str): 'relative' or 'absolute'

        Returns:
            (dict): a dict containing metadata with relative path

        """
        assert isinstance(data_in, dict)
        assert target in ['relative', 'absolute']

        data = deepcopy(data_in)

        # Convert absolute path to relative path
        if 'path' in data.keys():
            if isinstance(data['path'], str):
                data['path'] = self._solve_path(data['path'], target=target)
            elif isinstance(data['path'], list):
                for idx in range(len(data['path'])):
                    data['path'][idx] = self._solve_path(data['path'][idx], target=target)
            else:
                raise TypeError('Unsupported type')

        return data

    def _reindex(self):
        """Re-index."""
        indices = []

        for idx, value in enumerate(self._data.values()):
            # Count for self.__len__
            if self.orient in value.keys():
                if isinstance(value[self.orient], dict) or isinstance(value[self.orient], list):
                    indices += [[idx, i] for i in range(len(value[self.orient]))]
                else:
                    indices += [[idx, 0]]
            else:
                indices += [[idx, 0]]

        self._indices = indices
        self._indexed = True
        self._cursor = 0

    def __len__(self):
        """Return number of orients."""
        if not self._indexed:
            self._reindex()
        return len(self._indices)

    def __next__(self):
        """Return the next item."""
        data = super().__next__()

        return data

    def __getitem__(self, idx, **kwargs):
        """Return the corresponding item.

        Args:
            idx (int): Index of the item

        Returns:
            (dict): A dict of metadata

        """
        if not self._indexed:
            self._reindex()

        record_idx, orient_idx = self._indices[idx]
        data = super().__getitem__(record_idx, **kwargs)

        # Deserialize content
        data = self._solve_path_in_data(data, target='absolute')

        data = deepcopy(data)

        if self.orient in data.keys():
            if isinstance(data[self.orient], dict):
                data[self.orient] = \
                    {list(data[self.orient].keys())[orient_idx]:
                     list(data[self.orient].values())[orient_idx]}
            if isinstance(data[self.orient], list):
                data[self.orient] = [data[self.orient][orient_idx]]

        return data

    def add_data(self, data_in: dict, **kwargs):
        """Add data to db.

        Args:
            data_in (dict): data

        """
        data_to_store = self._solve_path_in_data(data_in, target='relative')
        super().add_data(data_to_store, **kwargs)
        self._indexed = False

    def remove_data(self, data):
        """Remove data from DB.

        Args:
            data (dict): data to remove

        """
        data_to_remove = self._solve_path_in_data(data, target='relative')
        super().remove_data(data_to_remove)
        self._indexed = False

    def read(self, *args, **kwargs):
        """Read function."""
        super().read(*args, **kwargs)
        self._indexed = False

    def save(self, *args, **kwargs):
        """Save function."""
        super().save(*args, **kwargs)
        self._database_id_db_handler.add_data({
            'database_id': self._database_id,
            'df_name': self._df_name
        })
        self._database_id_db_handler.save()

    @property
    def data(self):
        """Return data.

        Returns:
            (list): list of dicts

        """
        return [self._solve_path_in_data(data, target='absolute') for data in self._data.values()]

    @data.setter
    def data(self, data):
        """Setter for self.data.

        Args:
            data (list): new-data

        """
        assert isinstance(data, list)
        data_in = deepcopy(data)

        for value in data_in:
            value = self._solve_path_in_data(value, target='relative')
            self.add_data(value)

        self._indexed = False

    @property
    def _df_name(self):
        """Return _df_name.

        Returns:
            (str): name

        """
        template = self._config['_df_name'] \
            if '_df_name' in self._config.keys() else '{database_id}'
        digest_size = self._config['_hash_digest_size'] \
            if '_hash_digest_size' in self._config.keys() else 4
        database_id_hashed = hashlib.blake2s(
            self._database_id.encode('utf-8'),
            digest_size=digest_size
        ).hexdigest()
        return template.format(**{
            'database_id': database_id_hashed
        })

    @_df_name.setter
    def _df_name(self, value):
        """Setter for self._df_name."""
        raise RuntimeError(
            'Setting df_name is not supported in MetaDBHandler'
        )

    @property
    def _df(self):
        """Return df."""
        data = []
        for idx in range(len(self)):
            _data = self.__getitem__(idx, remove_internal_columns=False)
            assert isinstance(_data, dict)

            # Expand the contents of key `self.orient`
            if self.orient in _data.keys():
                if isinstance(_data[self.orient], list):
                    assert len(_data[self.orient]) == 1
                    if isinstance(_data[self.orient][0], dict):
                        value = flatten(next(iter(_data[self.orient])), reducer='dot')
                        _data = self._merger.merge(_data, value)
                    else:
                        _data[self.orient] = _data[self.orient][0]
                elif isinstance(_data[self.orient], dict):
                    assert len(_data[self.orient]) == 1
                    key = next(iter(_data[self.orient].keys()))
                    value = flatten(next(iter(_data[self.orient].values())), reducer='dot')
                    _data[self.orient] = key
                    _data = self._merger.merge(_data, value)
                else:
                    pass
            data.append(_data)

        df = self._df_from_dicts(data)
        return df

    @property
    def df(self):
        """Return df with display_names."""
        df = self._df
        self._to_display_names(df, inplace=True)
        return df


@register_handler(
    db_classes=['database_id'],
    db_engines=['tinydb', 'tinymongo', 'mongodb', 'montydb']
)
class DatabaseIDDBHandler(_BaseDBHandler):
    """Handler for database-id."""

    _df_class = 'database_id_df'
    _df_name = 'database_id_df'

    def _initialize_engine(self,
                           db_engine: Optional[str] = None,
                           db_host: Optional[str] = None,
                           db_name: Optional[str] = None,
                           db_username: Optional[str] = None,
                           db_password: Optional[str] = None):
        """Initialize DB engine."""
        # Load settings from environment variables
        engine = db_engine if db_engine is not None \
            else os.environ.get('PYDTK_META_DB_ENGINE', None)
        username = db_username if db_username is not None \
            else os.environ.get('PYDTK_META_DB_USERNAME', None)
        password = db_password if db_password is not None \
            else os.environ.get('PYDTK_META_DB_PASSWORD', None)
        host = db_host if db_host is not None \
            else os.environ.get('PYDTK_META_DB_HOST', None)
        database = db_name if db_name is not None \
            else os.environ.get('PYDTK_META_DB_DATABASE', None)

        super()._initialize_engine(engine, host, database, username, password)

    def remove_data(self, data, cascade=True):
        """Remove data from DB.

        Args:
            data (dict): Data to remove
            cascade (bool): If True, the corresponding table will also be removed

        """
        # Delete the corresponding table containing metadata
        if cascade:
            if 'df_name' in data.keys():
                target_collection = data['df_name']
                super().drop_table(target_collection)
            else:
                logging.warning(
                    'Skipped dropping the corresponding table '
                    'as key `df_name` is not included in the given data'
                )

        super().remove_data(data)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright Toolkit Authors

from copy import deepcopy
import hashlib
import logging
import os
from pathlib import Path

from tqdm import tqdm

from . import BaseDBHandler as _BaseDBHandler
from . import register_handler


@register_handler(db_classes=['meta'], db_engines=['sqlite', 'mysql', 'mariadb', 'postgresql'])
class MetaDBHandler(_BaseDBHandler):
    """Handler for metadb."""

    _database_id: str = ''
    _df_class = 'meta_df'
    _df_name = 'meta_df'

    def __init__(self, database_id='default', base_dir_path=None, **kwargs):
        """Initialize MetaDBHandler.

        Args:
            database_id (str): ID of the database (e.g. "Driving Behavior Database")
            base_dir_path (str): base directory for path

        """
        self._database_id = database_id
        super(MetaDBHandler, self).__init__(**kwargs)

        # Solve base-dir
        base_dir = base_dir_path
        if base_dir is None:
            if self._config.current_db['engine'] == 'sqlite':
                base_dir = os.path.dirname(self._config.current_db['host'])
            else:
                base_dir = '/'
        self.base_dir_path = os.path.realpath(base_dir)

        # Prepare another DB-handler for storing a list of database_id
        self._database_id_db_handler = DatabaseIDDBHandler(**{**kwargs, 'read_on_init': False})

    def _initialize_engine(self,
                           db_engine=None,
                           db_host=None,
                           db_name=None,
                           db_username=None,
                           db_password=None):
        """Initialize DB engine."""
        # Load settings from environment variables
        engine = db_engine if db_engine is not None else os.environ.get('PYDTK_META_DB_ENGINE', None)
        username = db_username if db_username is not None else os.environ.get('PYDTK_META_DB_USERNAME', None)
        password = db_password if db_password is not None else os.environ.get('PYDTK_META_DB_PASSWORD', None)
        host = db_host if db_host is not None else os.environ.get('PYDTK_META_DB_HOST', None)
        database = db_name if db_name is not None else os.environ.get('PYDTK_META_DB_DATABASE', None)

        super()._initialize_engine(engine, host, database, username, password)

    def _serialize_contents_and_solve_path(self, data_in):
        """Serialize contents and fix absolute path to relative one.

        Args:
            data_in (dict): a dict containing metadata

        Returns:
            (list): list of serialized dicts

        """
        assert 'contents' in data_in.keys()
        assert isinstance(data_in['contents'], dict)

        content_columns = self._config[self._df_class]['content_columns']

        # Convert absolute path to relative path
        if 'path' in data_in.keys():
            # noinspection PyTypeChecker
            data_path = Path(data_in['path'])
            try:
                relative_path = data_path.relative_to(self.base_dir_path)
                data_in['path'] = relative_path
            except ValueError as e:
                logging.warning('Could not resolve relative path to file: {}'.format(data_path))
                logging.warning(str(e))

        # Serialize contents
        data_out = []
        for content_name, content_info in data_in['contents'].items():
            for column_name in content_info.keys():
                if column_name not in content_columns:
                    raise KeyError('Unrecognized key "{0}" found in content "{1}".'
                                   '  Please check the config file'
                                   .format(column_name, content_name))
            data = deepcopy(data_in)
            data['contents'] = content_name
            data.update(content_info)
            data_out.append(data)

        return data_out

    def _deserialize_contents_and_solve_path(self, data_in):
        """Deserialize contents and fix relative path to absolute one.

        Args:
            data_in (dict): a dict containing metadata

        Returns:
            (dict): metadata with deserialized content

        """
        assert 'contents' in data_in.keys()
        assert isinstance(data_in['contents'], str)

        content_columns = self._config[self._df_class]['content_columns']

        data_out = data_in

        # Deserialize contents
        content_name = data_out['contents']
        data_out['contents'] = {content_name: {}}
        for column_name in content_columns:
            data_out['contents'][content_name][column_name] = data_out[column_name]
            del data_out[column_name]

        # Convert relative path to absolute path
        if 'path' in data_out.keys():
            data_out['path'] = os.path.join(self.base_dir_path, data_out['path'])

        return data_out

    def add_list_of_data(self, data_in, **kwargs):
        """Add listed data to db.

        Args:
            data_in (list): a list of dicts containing data

        """
        self.logger.info('(Preprocess) Pre-processing metadata')
        data_flat = []
        for data_item in tqdm(data_in, desc='Pre-process', leave=False):
            data_flat += self._serialize_contents_and_solve_path(data_item)
        super().add_list_of_data(data_flat, **kwargs)

    def __next__(self):
        """Return the next item."""
        data = super().__next__()

        # Deserialize content
        data = self._deserialize_contents_and_solve_path(data)

        return data

    def get_content_df(self):
        """Return content_df."""
        return self.content_df

    def get_file_df(self):
        """Return file_df."""
        return self.file_df

    def get_record_id_df(self):
        """Return record_id_df."""
        return self.record_id_df

    def save(self, *args, **kwargs):
        """Save function."""
        super().save(*args, **kwargs)
        self._database_id_db_handler.add_data({
            'database_id': self._database_id,
            'df_name': self._df_name
        })
        self._database_id_db_handler.save(remove_duplicates=True)

    @property
    def content_df(self):
        """Return content_df.

        Columns: record_id, path, content, msg_type, tag

        """
        df = self.df[['record_id', 'path', 'contents', 'msg_type', 'tags']]
        df = df.rename(columns={"contents": "content", "tags": "tag"})
        df['tag'] = df['tag'].apply(lambda x: list(set(x.split(';'))))
        df['path'] = df['path'].apply(lambda x: os.path.join(self.base_dir_path, x))
        return df

    @property
    def file_df(self):
        """Return file_df.

        Columns: path, record_id, type, content_type, start_timestamp, end_timestamp

        """
        df = self.df[['path', 'record_id', 'data_type', 'content_type', 'start_timestamp', 'end_timestamp']]
        df = df.rename(columns={"data_type": "type", "content_type": "content-type"})
        df = df.groupby(['path'], as_index=False).agg({
            'record_id': 'first',
            'type': 'first',
            'content-type': 'first',
            'start_timestamp': 'min',
            'end_timestamp': 'max',
        })
        df['path'] = df['path'].apply(lambda x: os.path.join(self.base_dir_path, x))
        return df

    @property
    def record_id_df(self):
        """Return record_id_df.

        Columns: 'record_id', 'duration', 'start_timestamp', 'end_timestamp', 'tags'

        """
        df = self.df[['record_id', 'start_timestamp', 'end_timestamp', 'tags']]
        df = df.groupby(['record_id'], as_index=False).agg({
            'start_timestamp': 'min',
            'end_timestamp': 'max',
            'tags': ':'.join
        })
        df['duration'] = df.end_timestamp - df.start_timestamp
        df['tags'] = df['tags'].apply(lambda x: list(set(x.split(';'))))
        return df

    @property
    def _df_name(self):
        """Return _df_name."""
        template = self._config[self._df_class]['df_name']
        database_id_hashed = hashlib.blake2s(
            self._database_id.encode('utf-8'),
            digest_size=self._config.hash.digest_size
        ).hexdigest()
        return template.format(**{
            'database_id': database_id_hashed
        })

    @_df_name.setter
    def _df_name(self, value):
        """Setter for self._df_name."""
        raise RuntimeError(
            'Setting df_name is not supported in MetaDBHandler'
        )


@register_handler(db_classes=['database_id'], db_engines=['sqlite', 'mysql', 'mariadb', 'postgresql'])
class DatabaseIDDBHandler(_BaseDBHandler):
    """Handler for database-id."""

    _df_class = 'database_id_df'
    _df_name = 'database_id_df'

    def _initialize_engine(self,
                           db_engine=None,
                           db_host=None,
                           db_name=None,
                           db_username=None,
                           db_password=None):
        """Initialize DB engine."""
        # Load settings from environment variables
        engine = db_engine if db_engine is not None else os.environ.get('PYDTK_META_DB_ENGINE', None)
        username = db_username if db_username is not None else os.environ.get('PYDTK_META_DB_USERNAME', None)
        password = db_password if db_password is not None else os.environ.get('PYDTK_META_DB_PASSWORD', None)
        host = db_host if db_host is not None else os.environ.get('PYDTK_META_DB_HOST', None)
        database = db_name if db_name is not None else os.environ.get('PYDTK_META_DB_DATABASE', None)

        super()._initialize_engine(engine, host, database, username, password)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright Toolkit Authors

"""DB Engines for V4DBHandler."""

from datetime import datetime
import logging
from typing import Optional

from tinydb import TinyDB, Query
from tinydb import __version__ as tinydb_version

logger = logging.getLogger(__name__)

DEFAULT_COLLECTION_NAME = 'default'


def connect(
    db_host: str,
    collection_name: Optional[str] = None,
    **kwargs
):
    """Connect to DB.

    Args:
        db_host (str): database host
        collection_name (str): collection name

    Returns:
        (any): connection

    """
    if collection_name is None:
        collection_name = DEFAULT_COLLECTION_NAME

    if tinydb_version.startswith('4'):
        db = TinyDB(db_host)
        db.default_table_name = collection_name
    elif tinydb_version.startswith('3'):
        db = TinyDB(db_host, default_table=collection_name)
    else:
        raise RuntimeError('TinyDB version < 3, >4 is not supported')
    return db


def read(db,
         query: Optional[dict or Query] = None,
         **kwargs):
    """Read data from DB.

    Args:
        db (TinyDB): DB connection
        query (dict or Query): Query to select items
        **kwargs: kwargs for function `pandas.read_sql_query`
                  or `influxdb.DataFrameClient.query`

    Returns:
        (list, int): list of data and total number of records

    """
    if query:
        data = db.search(query)
    else:
        data = db.all()

    return data, len(data)


def upsert(db, data, **kwargs):
    """Write data to DB.

    Args:
        db (TinyDB): DB connection
        data (list): data to save

    """
    for record in data:
        _record = _fix_datetime(record)
        uuid = _record['_uuid']
        db.upsert(_record, Query()._uuid == uuid)


def remove(db, uuids, **kwargs):
    """Remove data from DB.

    Args:
        db (TinyDB): DB connection
        uuids (list): A list of unique IDs

    """
    for uuid in uuids:
        db.remove(Query()._uuid == uuid)


def drop_table(db, name, **kwargs):
    """Drop a table from DB.

    Args:
        db (TinyDB): DB connection
        name (str): Name of the target table

    """
    if tinydb_version.startswith('4'):
        db.drop_table(name)
    else:
        db.purge_table(name)


def exist_table(db, name, **kwargs):
    """Check if the specified table (collection) exist.

    Args:
        db (TinyDB): DB connection
        name (str): Name of the target table

    """
    return name in list(db.tables())


def _fix_datetime(data: dict):
    """Fix datetime object.

    Args:
        data (dict): Input data

    Returns:
        (dict): Fixed data

    """
    _data = {}
    for key, value in data.items():
        if isinstance(value, datetime):
            _data[key] = value.timestamp()
        else:
            _data[key] = value

    return _data



#!/usr/bin/env python3

# Copyright Toolkit Authors

import pprint


class IO(object):
    """IO related operations."""

    @staticmethod
    def read(
        file: str,
        content: str = None,
    ):
        """Read a given file.

        Args:
            file (str): Path to a file
            content (str): Content in the file to read

        """
        from pydtk.io import BaseFileReader
        reader = BaseFileReader()
        timestamps, data, columns = reader.read(path=file, contents=content, as_ndarray=False)

        pprint.pprint(data)



from unittest import TestCase
from bson import SON
import pymongo
import pql

class PqlAggregationTest(TestCase):

    def compare(self, expression, expected):
        self.assertEqual(pql.AggregationParser().parse(expression), expected)

class PqlAggregationPipesTest(PqlAggregationTest):

    def test_match(self):
        self.assertEqual(pql.match('a == 1'), [{'$match': {'a': 1}}])
    
    def test_group(self):
        for group_func in pql.AggregationGroupParser.GROUP_FUNCTIONS:
            self.assertEqual(pql.group(_id='foo', total=group_func + '(bar)'),
                             [{'$group': {'_id': '$foo', 'total': {'$' + group_func: '$bar'}}}])

    def test_invalid_group(self):
        with self.assertRaises(pql.ParseError):
            pql.group(_id='foo', total='bar(1)')
        with self.assertRaises(pql.ParseError):
            pql.group(_id='foo', total='min(1, 2)')

    def test_project(self):
        self.assertEqual(pql.project(foo='bar', a='b + c'),
                         [{'$project': {'foo': '$bar', 'a': {'$add': ['$b', '$c']}}}])

    def test_skip(self):
        self.assertEqual(pql.skip(3), [{'$skip': 3}])

    def test_limit(self):
        self.assertEqual(pql.limit(2), [{'$limit': 2}])

    def test_unwind(self):
        self.assertEqual(pql.unwind('foo'), [{'$unwind': '$foo'}])

    def test_sort(self):
        self.assertEqual(pql.sort('a'), [{'$sort': SON([('a', pymongo.ASCENDING)])}])
        self.assertEqual(pql.sort(['a', '-b', '+c']),
                         [{'$sort': SON([('a', pymongo.ASCENDING),
                                         ('b', pymongo.DESCENDING),
                                         ('c', pymongo.ASCENDING)])}])

class PqlAggregationDataTypesTest(PqlAggregationTest):

    def test_bool(self):
        self.compare('True', True)
        self.compare('true', True)
        self.compare('False', False)
        self.compare('false', False)
        self.compare('None', None)
        self.compare('null', None)

class PqlAggregationSimpleProjectionTest(PqlAggregationTest):

    def test(self):
        self.compare('a', '$a')

    def test_nested(self):
        self.compare('a.b.c', '$a.b.c')

class PqlAggregationLogicTest(PqlAggregationTest):

    def test_and(self):
        self.compare('a and b', {'$and': ['$a', '$b']})

    def test_or(self):
        self.compare('a or b', {'$or': ['$a', '$b']})

    def test_not(self):
        self.compare('not a', {'$not': '$a'})

class PqlAggregationBoolTest(PqlAggregationTest):

    def test_cmp(self):
        self.compare('cmp(a, "bar")', {'$cmp': ['$a', 'bar']})

    def test_eq(self):
        self.compare('a == 0', {'$eq': ['$a', 0]})

    def test_gt(self):
        self.compare('a > 0', {'$gt': ['$a', 0]})

    def test_gte(self):
        self.compare('a >= 0', {'$gte': ['$a', 0]})

    def test_lt(self):
        self.compare('a < 0', {'$lt': ['$a', 0]})

    def test_lte(self):
        self.compare('a <= 0', {'$lte': ['$a', 0]})

    def test_ne(self):
        self.compare('a != 0', {'$ne': ['$a', 0]})

class PqlAggregationArithmicTest(PqlAggregationTest):
    
    def test_add(self):
        self.compare('a + 1', {'$add': ['$a', 1]})
        
    def test_divide(self):
        self.compare('a / 1', {'$divide': ['$a', 1]})

    def test_mod(self):
        self.compare('a % 1', {'$mod': ['$a', 1]})

    def test_multiply(self):
        self.compare('a * 1', {'$multiply': ['$a', 1]})

    def test_subtract(self):
        self.compare('a - 1', {'$subtract': ['$a', 1]})

class PqlAggregationStringTest(PqlAggregationTest):

    def test_concat(self):
        self.compare('concat("foo", "bar", b)', {'$concat': ['foo', 'bar', '$b']})

    def test_strcasecmp(self):
        self.compare('strcasecmp("foo", b)', {'$strcasecmp': ['foo', '$b']})

    def test_substr(self):
        self.compare('substr("foo", 1, 2)', {'$substr': ['foo', 1, 2]})

    def test_toLower(self):
        self.compare('toLower(a)', {'$toLower': ['$a']})

    def test_toUpper(self):
        self.compare('toUpper(a)', {'$toUpper': ['$a']})

class PqlAggregationDateTest(PqlAggregationTest):
    def test(self):
        for func in ['dayOfYear', 'dayOfMonth', 'dayOfWeek',
                     'year', 'month', 'week',
                     'hour', 'minute', 'second', 'millisecond']:
            self.compare('{0}(a)'.format(func), {'${0}'.format(func): ['$a']})

class PqlConditionTest(PqlAggregationTest):
    def test_if(self):
        self.compare('a if b > 3 else c', {'$cond': [{'$gt': ['$b', 3]}, '$a', '$c']})

    def test_if_null(self):
        self.compare('ifnull(a + b, 100)', {'$ifnull': [{'$add': ['$a', '$b']}, 100]})
    
class PqlAggregationSanityTest(PqlAggregationTest):
    def test(self):
        self.compare('a + b / c - 3 * 4 == 1',
                     {'$eq': [
                         {'$subtract': [{'$add': ['$a', {'$divide': ['$b', '$c']}]},
                                        {'$multiply': [3, 4]}]},
                         1]})

class PqlAggregationErrorsTest(PqlAggregationTest):
    def test_invalid_num_args(self):
        with self.assertRaises(pql.ParseError):
            self.compare('ifnull(1)', None)
        with self.assertRaises(pql.ParseError):
            self.compare('ifnull()', None)

    def test_invalid_func(self):
        with self.assertRaises(pql.ParseError):
            self.compare('foo(10)', None)

    def test_invalid_comparators(self):
        with self.assertRaises(pql.ParseError):
            self.compare('1 < a < 3', None)



"""Exceptions."""


class DatabaseNotInitializedError(BaseException):
    """Database is not initialized."""

    pass


class InvalidDatabaseConfigError(BaseException):
    """Database-config is not valid."""

    pass



#!/usr/bin/env python3

# Copyright Toolkit Authors (Yusuke Adachi)

import argparse
from collections import OrderedDict
import json
import logging
import os
import re
import rosbag


def _read_rosbag(file):
    info = {}
    with rosbag.Bag(file, "r") as bag:
        info["start_time"] = bag.get_start_time()
        info["end_time"] = bag.get_end_time()
        # get topic names
        info["topic_info"] = bag.get_type_and_topic_info()
    return info


def _read_timestamp(file):
    """Get start and end time from timestamp csv file."""
    try:
        with open(file, 'r') as f:
            rows = f.readlines()
            starttime, endtime = float(rows[0].split(",")[0]), float(rows[-1].split(",")[0])
        starttime, endtime = starttime / (10**3), endtime / (10**3)
    except IOError:
        starttime, endtime = "Nan", "Nan"
    return starttime, endtime


def _get_time_info(file):
    """Get start and end time."""
    if file.endswith(".mp4"):
        file = file.replace(".mp4", "_timestamps.csv")
        starttime, endtime = _read_timestamp(file)
    elif file.endswith(".bag"):
        bag_info = _read_rosbag(file)
        starttime, endtime = bag_info["start_time"], bag_info["end_time"]
    elif file.endswith("timestamps.csv"):
        starttime, endtime = _read_timestamp(file)
    else:
        starttime, endtime = "no info", "no info"
    return starttime, endtime


def _check_camera_contents(file):
    # camera_id = int(os.path.basename(file).split("_")[1])
    camera_id = re.split('[._]', os.path.basename(file))
    camera_id = int(camera_id[1])

    camera_info_dict = [
        "no info",
        "camera/front-center",
        "camera/front-left",
        "camera/front-right",
        "camera/rear-center",
        "camera/rear-left",
        "camera/rear-right",
        "camera/driver-face",
        "camera/driver-foot",
        "camera/driver-face-black_and_white"
    ]
    contents = camera_info_dict[camera_id]
    tags = re.split('[/-]', contents)
    return contents, tags


def _check_contents(file):
    contents = OrderedDict()
    if file.endswith("aac"):
        contents["microphone"] = {"tags": ["microphone", "sound"]}
    elif os.path.basename(file).startswith("camera"):
        contents_name, contents_tags = _check_camera_contents(file)
        if file.endswith("mp4"):
            contents_tags.append("image")
        elif file.endswith("timestamps.csv"):
            contents_tags.append("timestamps")
        contents[contents_name] = {"tags": contents_tags}
    elif file.endswith("bag"):
        bag_info = _read_rosbag(file)
        topic_info = bag_info["topic_info"]
        topics = [topic for topic in topic_info[1].keys()]
        topics.sort()
        for topic in topics:
            if topic.startswith("/mobileye"):  # mobileye data is not supported at this version
                continue
            contents[topic] = {}
            contents[topic]["msg_type"] = topic_info[1][topic].msg_type
            contents[topic]["msg_md5sum"] = topic_info[0][topic_info[1][topic].msg_type]
            contents[topic]["count"] = topic_info[1][topic].message_count
            contents[topic]["frequency"] = topic_info[1][topic].frequency
            contents[topic]["tags"] = re.split('[_/-]', topic[1:])
    return contents


def _check_content_type(file):
    type_dict = {
        "aac": "audio/aac",
        "csv": "text/csv",
        "mp4": "video/mp4",
        "bag": "application/rosbag"
    }
    extention = file.split(".")[-1]
    try:
        content_type = type_dict[extention]
    except KeyError:
        content_type = "unknown"
    return content_type


def dump_json(file_path, json_path):
    """Write file information in json."""
    logging.debug("Loading %s" % file_path)
    attributes = OrderedDict()

    attributes["description"] = "Driving Database"
    attributes["database_id"] = "Driving Behavior Database"
    attributes["record_id"] = file_path.split("/")[-3]
    attributes["type"] = "raw_data"  # 収録データ
    attributes["path"] = file_path
    attributes["start_timestamp"], attributes["end_timestamp"] = _get_time_info(file_path)

    attributes["content-type"] = _check_content_type(file_path)
    attributes["contents"] = _check_contents(file_path)

    with open(json_path, mode="w") as f:
        json.dump(attributes, f, ensure_ascii=False, indent=4, separators=(',', ': '))
    logging.debug("Finish writing: %s" % json_path)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="Make json file for NU data tool")
    parser.add_argument('file_path', help='Path to drive ID  directory')
    parser.add_argument('--json_path', default="", help='Path to an output json file')
    parser.add_argument('-v', '--verbose', action='store_true', help='Verbose mode')
    args = parser.parse_args()
    json_path = args.json_path if not args.json_path == "" else args.file_path + ".json"

    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    dump_json(args.file_path, json_path)



#!/usr/bin/env python3

# Copyright Toolkit Authors (Yusuke Adachi)

import argparse
import pandas as pd
import os
import pickle
import shutil


def concat_df(pkl_files, out_pkl):
    """Concatinate some pickle files.

    Args:
        pkl_files (list): List of picke files to concatinate.
        out_pkl (str): Output pickle file.

    """
    for i, pkl in enumerate(pkl_files):
        with open(pkl, "rb") as f:
            df_dict = pickle.load(f)
            if i == 0:
                file_df = df_dict['file_df']
                content_df = df_dict['content_df']
                record_id_df = df_dict['record_id_df']
            else:
                file_df = pd.concat([file_df, df_dict['file_df']])
                content_df = pd.concat([content_df, df_dict['content_df']])
                record_id_df = pd.concat([record_id_df, df_dict['record_id_df']])

    if os.path.isfile(out_pkl):
        shutil.move(out_pkl, out_pkl + '.bak')
    with open(out_pkl, "wb") as f:
        pickle.dump({"file_df": file_df,
                     "content_df": content_df,
                     "record_id_df": record_id_df},
                    f, protocol=2)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="This script is bababa")
    parser.add_argument('in_pkls', help='Pickle files to concatinate.', nargs='*')
    parser.add_argument('--out_pkl', help='Output pickle file.')
    args = parser.parse_args()
    pkl_list, out_pkl = args.in_pkls, args.out_pkl

    concat_df(pkl_list, out_pkl)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright Toolkit Authors

"""CAN Decoder."""

import os
import csv
import sys
import re
import logging
import bitstring
import time
import datetime
import argparse


class BitAssign(object):
    """A class for storing assignment information of a single bit."""

    def __init__(self, data_label, data_name, can_id, data_position, data_length,
                 minimum_renewal_time, minimum_renewal_time_comment, unit, resolution,
                 resolution_comment, offset_physical, offset_cpu, signed):
        """Initialize BitAssign.

        Args:
            data_label (str): Data label
            data_name (str): Data name
            can_id (str): ID of CAN
            data_position (str): The first position of the corresponding bits
            data_length (str): Number of bits corresponding to the data
            minimum_renewal_time (str): Optional
            minimum_renewal_time_comment (str): Optional
            unit (str): Unit of the data
            resolution (str): Resolution of the data
            resolution_comment (str): Optional
            offset_physical (str): Offset to add after converting to a physical value
            offset_cpu (str): Offset to add before converting to a physical value
            signed (str): Whether the data is signed or not

        """
        self.data_label = data_label.replace(" ", "")
        self.data_name = data_name
        self.can_id = can_id.replace(" ", "").lower()
        self.data_position = data_position.replace(" ", "")
        self.data_length = data_length.replace(" ", "")
        self.minimum_renewal_time = minimum_renewal_time.replace(" ", "")
        self.minimum_renewal_time_comment = minimum_renewal_time_comment
        self.unit = unit.replace(" ", "")
        self.resolution = resolution.replace(" ", "")
        self.resolution_comment = resolution_comment
        self.offset_physical = offset_physical.replace(" ", "")
        self.offset_cpu = offset_cpu.replace(" ", "")
        self.signed = signed

        self.reformat()

    def reformat(self):
        """Remove unregular characters from bit assign information."""
        self.data_position = re.sub("[^0-9|.|-]", "", self.data_position)
        self.data_length = re.sub("[^0-9|.|-]", "", self.data_length)
        self.minimum_renewal_time = re.sub("[^0-9|.|-]", "", self.minimum_renewal_time)
        self.resolution = re.sub("[^0-9|.|-]", "", self.resolution)
        self.offset_physical = re.sub("[^0-9|.|-]", "", self.offset_physical)

        if self.resolution == '' or self.resolution == '-':
            self.resolution = '1.0'

        if self.offset_physical == '' or self.offset_physical == '-':
            self.offset_physical = '0'


class BitAssignInfo(object):
    """A class for handling CAN assignment list."""

    def __init__(self, path_to_assign_list=None):
        """Initialize BitAssignInfo."""
        self.logger = logging.getLogger(type(self).__name__)
        self.signal_list = {}
        self.path_to_assign_list = path_to_assign_list
        if path_to_assign_list is not None:
            self.load_from_csv(path_to_assign_list)

    def load_from_csv(self, path_to_assign_list):
        """Load can bit assign from csv.

        Args:
            path_to_assign_list (str): Path to csv

        """
        # Check
        if path_to_assign_list == "":
            self.logger.warning("empty path is given")
            return

        # Check if the file exists
        if not os.path.exists(path_to_assign_list):
            self.logger.warning("file does not exists: {}".format(path_to_assign_list))
            return

        # Prepare
        signal_list = {}

        # Load the file
        with open(path_to_assign_list, 'rU') as f:
            reader = csv.reader(f)
            _ = next(reader)  # Header
            for row in reader:
                bit_assign_info = BitAssign(*row)
                signal_list.update({row[0]: bit_assign_info})

        self.signal_list = signal_list
        self.logger.debug("loaded {0} signal assignment from file: {1}"
                          .format(len(signal_list), path_to_assign_list))

        if len(signal_list) == 0:
            self.logger.error("Could not load bit assign list: {}".format(path_to_assign_list))
            sys.exit(1)

    def bit_assigns_from_can_id(self, can_id):
        """Returns list of bit assign information which corresponds to the given CAN ID.

        Args:
            can_id (str): ID of CAN

        """
        # Check
        if len(self.signal_list) == 0:
            if self.path_to_assign_list != "":
                self.load_from_csv(self.path_to_assign_list)
            else:
                self.logger.error("please set the path to can bit assign list")

        # Find corresponding bit assign and return
        bit_assigns = []
        for bit_assign in self.signal_list.values():
            # self.logger.debug("can_id (assign): {0}, can_id (data): {1}"
            #                   .format(bit_assign.can_id, can_id.replace(" ", "")))
            if can_id.replace(" ", "") == bit_assign.can_id:
                bit_assigns.append(bit_assign)
        return bit_assigns


class CANData(object):
    """A class for handling the decoded CAN signals."""

    def __init__(self, time_in_nsec, data_label, data):
        """Initialize CanData.

        Args:
            time_in_nsec (str): Time in nano-seconds
            data_label (str): Data label
            data (str): Data

        """
        self.time_in_nsec = time_in_nsec
        self.time_in_datetime = datetime.datetime.fromtimestamp(float(time_in_nsec) / 1.0E9)
        self.data_label = data_label
        self.data = data


class CANDecoder(object):
    """CAN csv file decoder."""

    def __init__(self, path_to_assign_list=None):
        """Initialize CANDecoder.

        Args:
            path_to_assign_list (str): Path to bit-assign info file (csv)

        """
        self.logger = logging.getLogger(type(self).__name__)
        self.path_to_assign_list = path_to_assign_list
        self.path_to_csv = ""
        self.bit_assign_info = BitAssignInfo(path_to_assign_list)
        self.data = []
        self.f = None
        self.reader = None

    def load_bit_assign_list(self, path_to_assign_list):
        """Load bit-assign list from a file.

        Args:
            path_to_assign_list (str): Path to the file

        """
        self.path_to_assign_list = path_to_assign_list
        self.bit_assign_info = BitAssignInfo(path_to_assign_list)

    def open_csv(self, path_to_csv=""):
        """Open a csv file and prepare file-pointer.

        Args:
            path_to_csv (str): Path to the csv file containing CAN data

        """
        if path_to_csv == "":
            path_to_csv = self.path_to_csv
        else:
            self.path_to_csv = path_to_csv

        # Check file existence
        if not os.path.exists(path_to_csv):
            self.logger.warning("file does not exists: {}".format(path_to_csv))
            sys.exit(1)

        f = open(path_to_csv, 'rU')
        reader = csv.reader(f)
        _ = next(reader)  # Skip header
        self.f = f
        self.reader = reader

    def close_csv(self):
        """Close the file."""
        self.f.close()
        self.reader = None

    def analyze_line(self):
        """Returns list of CANData in current line.

        Returns:
            (list): A list of CANData

        """
        # check
        if self.f is None:
            self.logger.error("file is not opened yet")
            return None

        if self.f.closed:
            self.logger.error("file is already closed")
            return None

        # Prepare
        data = []

        # Get line
        row = next(self.reader)
        if not row:
            raise EOFError

        # Analyze
        for bit_assign in self.bit_assign_info.bit_assigns_from_can_id(row[1]):
            data.append(
                CANData(row[0], bit_assign.data_label, self.unpack_data(row[3], bit_assign)))

        return data

    def analyze_csv(self, path_to_csv="", start_time=None, end_time=None):
        """Returns list of CANData.

        Args:
            path_to_csv (str): Path to a csv file containing CAN data
            start_time (datetime): Start-time to extract data
            end_time (datetime): End-time to extract data

        Returns:
            (list): A list of CANData

        """
        if path_to_csv == "":
            path_to_csv = self.path_to_csv
        else:
            self.path_to_csv = path_to_csv

        # check file existence
        if not os.path.exists(path_to_csv):
            self.logger.warning("file does not exists: {}".format(path_to_csv))
            return None

        # prepare
        data = []
        line_index = 1
        line_datetime = None

        # start analyzing
        self.logger.debug("started loading csv file")
        with open(path_to_csv, 'rU') as f:
            reader = csv.reader(f)
            _ = next(reader)    # Header

            for row in reader:
                timestamp = row[2]
                can_id = bitstring.ConstBitArray(uint=int(row[5]), length=32).hex.lstrip(
                    "0").lower()
                _ = int(row[6])     # can_length
                can_data = "".join(
                    [bitstring.ConstBitArray(uint=int(b), length=8).hex for b in row[7:15]])
                for bit_assign in self.bit_assign_info.bit_assigns_from_can_id(can_id):
                    data.append(CANData(timestamp, bit_assign.data_label,
                                        self.unpack_data(can_data, bit_assign)))
                if line_index % 500000 == 0:
                    self.logger.debug("{} lines decoded".format(line_index))
                line_index += 1

        # check
        if line_datetime is not None and line_datetime < end_time:
            self.logger.error(
                "This csv file ends {0} before end_time {1}".format(line_datetime, end_time))
            return []

        # return
        return data

    @staticmethod
    def unpack_data(data_string, bit_assign):
        """Decodes data_string to signal value.

        Args:
            data_string (str): Data string
            bit_assign (BitAssign): A BitAssign object

        Returns:
            (str): Decoded value

        """
        c = bitstring.ConstBitArray(hex=data_string.replace(" ", ""))
        if "." in bit_assign.data_position:
            data_position_1 = int(bit_assign.data_position.split(".")[0])
            data_position_2 = int(bit_assign.data_position.split(".")[1])
        else:
            data_position_1 = int(bit_assign.data_position)
            data_position_2 = 0
        start_position = data_position_1 * 8 - data_position_2
        data_len = int(bit_assign.data_length)  # bits

        data = bitstring.ConstBitArray(
            bin=c.bin[(start_position - data_len):start_position])

        if bit_assign.signed == '1':
            base_value = data.int
        else:
            base_value = data.uint

        value = base_value * float(bit_assign.resolution) + float(bit_assign.offset_physical)

        return value


class CANDeserializer(object):
    """Deserialize CAN data."""

    def __init__(self, signal_list=None):
        """Initialize CANDeserializer.

        Args:
            signal_list (list): A list of signals (str) to deserialize

        """
        self.logger = logging.getLogger(type(self).__name__)
        self.signal_list = ["brake_pos", "turn_signal"] if signal_list is None else signal_list

    def deserialize(self, can_data, start_time=None, end_time=None, fps=5):
        """Deserializes given CAN data from start_time to end_time with given fps.

        Args:
            can_data (list): A list of CANData objects
            start_time (datetime): Start-time to deserialize data
            end_time (datetime): End-time to deserialize data
            fps (int): Fps to deserialize

        Returns:
            (dict): Deserialized data

        """
        # check
        if len(can_data) == 0:
            self.logger.error("empty can data is given")
            return None

        if start_time is None:
            start_time = can_data[0].time_in_datetime
            self.logger.debug("setting start_time as: {}".format(start_time))

        if end_time is None:
            end_time = can_data[-1].time_in_datetime
            self.logger.debug("setting end_time as: {}".format(end_time))

        # prepare
        current_time_start = start_time
        current_time_end = start_time + datetime.timedelta(seconds=(float(1.0) / float(fps)))
        current_index = 0
        # end_of_data_flag = False

        # prepare tables
        previous = {signal: None for signal in self.signal_list}
        result = {signal: [] for signal in self.signal_list}
        result["time_in_datetime"] = []

        # loop
        self.logger.debug("started can data deserialization")
        while current_index < len(can_data):

            # initialize signal_temp dictionary
            signal_temp = {signal: [] for signal in self.signal_list}
            signal_temp_previous = {signal: [] for signal in self.signal_list}

            # store can signals
            for can_signal in can_data[current_index:]:
                if current_time_start <= can_signal.time_in_datetime < current_time_end:
                    if can_signal.data_label in self.signal_list:
                        signal_temp[can_signal.data_label].append(can_signal.data)
                    current_index += 1
                else:
                    if can_signal.time_in_datetime < current_time_start:
                        if can_signal.data_label in self.signal_list:
                            signal_temp_previous[can_signal.data_label].append(can_signal.data)
                        current_index += 1
                    elif can_signal.time_in_datetime > current_time_end:
                        break

            # calculate average and store it
            for signal in self.signal_list:
                # store previous signals if exists
                if len(signal_temp_previous[signal]) != 0:
                    previous[signal] = sum(signal_temp_previous[signal]) / float(
                        len(signal_temp_previous[signal]))

                # store current signals
                if len(signal_temp[signal]) == 0:
                    if len(result[signal]) == 0:
                        if previous[signal] is None:
                            result[signal].append(None)
                        else:
                            result[signal].append(previous[signal])
                    else:
                        result[signal].append(result[signal][-1])
                else:
                    result[signal].append(
                        sum(signal_temp[signal]) / float(len(signal_temp[signal])))

            # store time
            result["time_in_datetime"].append(current_time_start)

            # break
            if current_index >= len(can_data) \
                    or can_data[current_index].time_in_datetime > end_time:
                self.logger.debug("escaping")
                break

            # next step
            current_time_start = current_time_end
            current_time_end = current_time_end + datetime.timedelta(
                seconds=(float(1.0) / float(fps)))

        # return
        return result


def main(args):
    """Main function.

    Args:
        args (ArgumentParser): arguments

    """
    # Logging
    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.WARNING)

    can_decoder = CANDecoder(args.assign_list)
    can_data = can_decoder.analyze_csv(args.path_to_csv)

    if args.verbose:
        for data in can_data:
            print("{0}, {1}, {2}".format(data.unix_time_in_millisecond, data.data_label, data.data))

    can_deserializer = CANDeserializer()
    signal_list = can_deserializer.signal_list
    can_data_deserialized = can_deserializer.deserialize(can_data, fps=10)

    # print
    print("{0:>20},{1}".format("Timestamp", ",".join(["{0:>16}".format(k) for k in signal_list])))
    for idx, stamp in enumerate(can_data_deserialized['time_in_datetime']):
        print(
            "{0:>20},{1}".format(str(int(time.mktime(stamp.timetuple()))) + stamp.strftime("%f000"),
                                 ",".join(
                                     ["{0:>16,.03f}".format(can_data_deserialized[k][idx]) for k in
                                      signal_list])))


if __name__ == '__main__':
    # parser
    parser = argparse.ArgumentParser(
        description="Decodes a CAN signal csv file according to a given bit assign list")
    parser.add_argument('path_to_csv')
    parser.add_argument('-a', '--assign-list', type=str, default='./bit_assign_list.csv',
                        help='Path to bit assing list (csv file)')
    parser.add_argument('-v', '--verbose', action='store_true', help='Verbose mode')
    args = parser.parse_args()

    main(args)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright Toolkit Authors

"""Dummy preprocessing function."""

from .preprocess import BasePreprocess


class PassThrough(BasePreprocess):
    """Pass through processing."""

    def __init__(self):
        super(PassThrough, self)

    def processing(self, timestamps, values):
        """No processing.

        Args:
            timestamps ():
            values (ndarray): Signal sequence.

        Return: Same as input.

        """
        return timestamps, values


class AddBias(BasePreprocess):
    """Add bias processing."""

    def __init__(self, bias):
        super(AddBias, self)
        self.bias = bias

    def processing(self, timestamps, values):
        """Add constant bias.

        Args:
            timestamps ():
            values (ndarray): Signal sequence.

        Return: Same as input.

        """
        return timestamps, values + self.bias



#!/usr/bin/env python3

# Copyright Toolkit Authors

import fire
import logging
import os
from pathlib import Path
import time

from tqdm import tqdm

from pydtk.db import V3DBHandler, V4DBHandler
from pydtk.db.v4.engines import DB_ENGINES
from pydtk.models import MetaDataModel


def _find_json(db_dir):
    """Search with tags.

    Args:
        db_dir (str): Directory path which has database files.
    Returns:
        (list): List of json file path.

    """
    p = Path(db_dir)
    json_list = list(p.glob("**/*.json"))
    json_list.sort()
    return json_list


def main(target_dir,
         database_id='default',
         base_dir=None,
         output_db_engine=None,
         output_db_host=None,
         output_db_username=None,
         output_db_password=None,
         output_db_name=None,
         verbose=False):
    """Create meta_db.

    Args:
        target_dir (str): Path to database directory
        database_id (str): ID of the database (e.g. "Driving Behavior Database")
        base_dir (str): Directory path where each file-path will be based on
        output_db_engine (str): Database engine
        output_db_host (str): HOST of database
        output_db_username (str): Username for the database
        output_db_password (str): Password for the database
        output_db_name (str): Database name
        verbose (bool): Verbose mode

    """
    if verbose:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    # Check
    if not os.path.isdir(target_dir):
        raise IOError('No such directory: {}'.format(target_dir))

    # Select DB-Handler
    db_engine = os.environ.get('PYDTK_META_DB_ENGINE', None)
    if output_db_engine is not None:
        db_engine = output_db_engine
    if db_engine is None:
        DBHandler = V4DBHandler
    else:
        if db_engine in DB_ENGINES.keys():
            DBHandler = V4DBHandler
        else:
            DBHandler = V3DBHandler

    # Search metadata files
    t0 = time.time()
    logging.info("Searching json files...")
    json_list = _find_json(target_dir)
    t1 = time.time()
    logging.info("Found {0} files.({1:.03f} secs)".format(len(json_list), t1 - t0))

    # Preparation
    base_dir_path = base_dir if base_dir is not None else target_dir
    handler = DBHandler(
        db_class='meta',
        db_engine=output_db_engine,
        db_host=output_db_host,
        db_username=output_db_username,
        db_password=output_db_password,
        db_name=output_db_name,
        database_id=database_id,
        base_dir_path=base_dir_path
    )

    # Append metadata to db
    logging.info('Loading metadata...')
    for path in tqdm(json_list, desc='Load metadata', leave=False):
        if not MetaDataModel.is_loadable(path):
            logging.warning('Failed to load metadata file: {}, skipped'.format(path))
            continue
        metadata = MetaDataModel()
        metadata.load(path)
        handler.add_data(metadata.data)
    t2 = time.time()
    logging.info("Finished loading metadata.({0:.03f} secs)".format(t2 - t1))

    # Export
    logging.info('Saving DB file...')
    handler.save(remove_duplicates=True)
    t3 = time.time()
    logging.info("Finished saving DB file.({0:.03f} secs)".format(t3 - t2))

    # Display
    logging.info("Done.(Total: {0:.03f} secs)".format(t3 - t0))


def script():
    """Function for tool.poetry.scripts."""
    fire.Fire(main)


if __name__ == '__main__':
    fire.Fire(main)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright Toolkit Authors

"""pydtk modules."""

from pydtk.io.writer import BaseFileWriter
from pydtk.io.reader import BaseFileReader
from pydtk.io.errors import NoModelMatchedError



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright Toolkit Authors

"""pydtk modules."""

from abc import ABCMeta
import numpy as np
import os

from pydtk.models import MODELS_BY_PRIORITY, MetaDataModel
from pydtk.preprocesses import PassThrough
from pydtk.io.errors import NoModelMatchedError


class BaseFileReader(metaclass=ABCMeta):
    """Base file reader."""

    _model = None   # model used for loading a file

    def __init__(self, **kwargs):
        self.preprocesses = [PassThrough()]

    @classmethod
    def _select_model(cls, file_metadata):
        """Select a proper model based on the given file-metadata.

        Args:
            file_metadata (object): an MetaDataModel object

        """
        priorities = MODELS_BY_PRIORITY.keys()
        for priority in sorted(priorities, reverse=True):
            for model in MODELS_BY_PRIORITY[priority]:
                if model.is_loadable(**file_metadata.data):
                    return model
        raise NoModelMatchedError('No suitable model found for loading data: {}'.
                                  format(file_metadata))

    @property
    def model(self):
        """Return model."""
        return self._model

    @model.setter
    def model(self, model):
        self._model = model

    def add_preprocess(self, preprocess):
        """Add preprocessing function."""
        self.preprocesses += [preprocess]

    def read(self,
             metadata=None,
             as_generator=False,
             model_kwargs=None,
             as_ndarray=True,
             **kwargs
             ):
        """Read a file which corresponds to the given metadata.

        Args:
            metadata (MetaDataModel or dict): metadata of the data to load
            as_generator (bool): load data as a generator.
            model_kwargs (dict): kwargs to pass to the selected model

        Kwargs:
            path (str): path to a file
            contents (str or dict): content to load
            start_timestamp (float): start-timestamp
            end_timestamp (float): end-timestamp

        Returns:
            (object): an object of the corresponding model

        """
        if model_kwargs is None:
            model_kwargs = {}
        if metadata is None:
            if 'path' not in kwargs.keys():
                raise ValueError('Either metadata or path must be specified')
            # Look for the corresponding metadata file
            for ext in MetaDataModel._file_extensions:
                metadata_filepath = kwargs['path'] + ext
                if os.path.isfile(metadata_filepath):
                    metadata = MetaDataModel()
                    metadata.load(metadata_filepath)
            if metadata is None:
                raise IOError('Could not find metadata file')
        else:
            metadata = MetaDataModel(metadata)

        # Replace 'contents' in metadata to specify which content to load
        contents = metadata.data['contents'] if 'contents' in metadata.data.keys() else None
        if 'contents' in kwargs.keys():
            if isinstance(kwargs['contents'], dict):
                contents = kwargs['contents']
            if isinstance(kwargs['contents'], str):
                contents = next(iter([{k: v for k, v in contents.items()
                                       if k == kwargs['contents']}]))
            if len(contents) == 0:
                raise ValueError('No corresponding contents exist')

        # Replace other attributes with the given arguments
        metadata.data.update(kwargs)
        metadata.data.update({'contents': contents})

        # Select a suitable model and load data
        self.model = self._select_model(metadata)
        self.model = self.model(metadata=metadata, **model_kwargs)

        if as_generator:
            def load_sample_wise():
                for sample in self.model.load(as_generator=as_generator):
                    # Parse data
                    timestamp = np.array(sample['timestamps'])
                    data = np.array(sample['data'])
                    columns = self.model.columns
                    yield timestamp, data, columns
            return load_sample_wise()
        else:
            self.model.load()

            # Parse data
            timestamps = self.model.timestamps
            if as_ndarray:
                data = self.model.to_ndarray()
            else:
                data = self.model.data

            # Apply pre-processes
            for preprocess in self.preprocesses:
                timestamps, data = preprocess.processing(timestamps, data)

            columns = self.model.columns
            return timestamps, data, columns



#!/usr/bin/env python3

# Copyright Toolkit Authors

import os

from pydtk.db import DBHandler


class STATUS(object):
    """Status checker."""

    def environment(self):
        """Display environment variables."""
        env_vars = [
            "PYDTK_META_DB_ENGINE",
            "PYDTK_META_DB_USERNAME",
            "PYDTK_META_DB_PASSWORD",
            "PYDTK_META_DB_HOST",
            "PYDTK_META_DB_DATABASE",
        ]
        for env_var in env_vars:
            print(f"{env_var}:\t{os.environ.get(env_var, None)}")

    def env(self):
        """Display environment variables."""
        self.environment()

    def access(self):
        """Check access."""
        handler = DBHandler(
            db_class='database_id'
        )
        try:
            handler.read()
        except Exception:
            print("Can't access database: database_id")
            return
        # Check reading
        for row in handler:
            database_id = row["database_id"]
            readable = self._read_check(database_id=database_id)
            print(f'{database_id}\tread: {"O.K." if readable else "N.G."}')

    def _read_check(self,
                    database_id: str,
                    base_dir: str = '/'
                    ):
        """Check access to DB."""
        handler = DBHandler(
            db_class='meta',
            database_id=database_id,
            base_dir_path=base_dir,
            orient='contents'
        )
        try:
            handler.read(limit=1)
            return True
        except Exception:
            return False



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright Toolkit Authors

from abc import ABC

import cv2
import os

from pydtk.models import BaseModel, register_model


@register_model(priority=1)
class GenericImageModel(BaseModel, ABC):
    """A generic model for a image file."""

    _content_type = None  # allow any content-type
    _data_type = None   # allow any data-type
    _file_extensions = ['.png', '.jpg']
    _contents = None

    def __init__(self, **kwargs):
        super(GenericImageModel, self).__init__(**kwargs)

    def _load(self, path, **kwargs):
        """Load a image file.

        Args:
            path (str): path to a image file

        """
        data = cv2.imread(path)
        self.data = data

    def _save(self, path, **kwargs):
        """Save ndarray data to a image file.

        Args:
            path (str): path to the output image file

        """
        cv2.imwrite(path, self.data)

    def to_ndarray(self):
        """Return data as ndarray."""
        return self.data

    @classmethod
    def generate_contents_meta(cls, path, content_key='image'):
        """Generate contents metadata.

        Args:
            path (str): File path
            content_key (str): Key of content

        Returns:
            (dict): contents metadata

        """
        # Load file
        data = cv2.imread(path)
        shape = data.shape
        _, ext = os.path.splitext(path)

        # Generate metadata
        contents = {content_key: {'size': shape, 'tags': ['image', ext.replace('.', '')]}}

        return contents

    @property
    def timestamps(self, path):
        """Return timestamps as ndarray."""
        raise NotImplementedError

    @classmethod
    def generate_timestamp_meta(cls, path):
        """Generate contents metadata.

        Args:
            path (str): File path

        Returns:
            (list): [start_timestamp, end_timestamp]

        """
        raise NotImplementedError



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright Toolkit Authors

"""V3DB."""

# Handlers
from .handlers import BaseDBHandler as DBHandler
from .handlers.meta import MetaDBHandler
from .handlers.time_series import TimeSeriesDBHandler
from .handlers.time_series import TimeSeriesCassandraDBHandler
from .handlers.statistics import StatisticsDBHandler
from .handlers.statistics import StatisticsCassandraDBHandler

# Search engines
from .search_engines import BaseDBSearchEngine as DBSearchEngine
from .search_engines.time_series import TimeSeriesCassandraDBSearchEngine



#!/usr/bin/env python3

# Copyright Toolkit Authors

import logging
import re
import sys

import fire


def _call_as_subcommand(component):
    fire.Fire(component, command=sys.argv[2:], name=sys.argv[1])


def _check_pep515(args):
    """Check if argv contains `<number>_<number>`.

    Args:
        args (List[str]): input arguments

    """
    match = re.compile(r'^[0-9]+(_[0-9]+)+$')
    for arg in args:
        if match.fullmatch(arg) is not None:
            raise ValueError(
                f'Invalid value "{arg}": values with format "<number>_<number>" is not allowed'
            )


class CLI(object):
    """CLI of PyDTK.

    PyDTK is a toolkit for managing, retrieving, and processing data
    """

    @staticmethod
    def db():
        """Database related operations."""
        from pydtk.bin.sub_commands.db import DB
        _call_as_subcommand(DB)

    @staticmethod
    def model():
        """Model related operations."""
        from pydtk.bin.sub_commands.model import Model
        _call_as_subcommand(Model)

    @staticmethod
    def io():
        """IO related operations."""
        from pydtk.bin.sub_commands.io import IO
        _call_as_subcommand(IO)

    @staticmethod
    def version():
        """Show PyDTK's version."""
        import pydtk
        print(f'Version: {pydtk.__version__}')
        print(f'Commit ID: {pydtk.__commit_id__}')

    @staticmethod
    def status():
        """Status check operations."""
        from pydtk.bin.sub_commands.status import STATUS
        _call_as_subcommand(STATUS)


def script():
    """Function for tool.poetry.scripts."""
    verbose = False
    if '-v' in sys.argv:
        verbose = True
        del sys.argv[sys.argv.index('-v')]
    if '--verbose' in sys.argv:
        verbose = True
        del sys.argv[sys.argv.index('--verbose')]

    if verbose:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.ERROR)
        sys.tracebacklimit = 0

    # Check args
    _check_pep515(sys.argv)

    if len(sys.argv) > 1:
        fire.Fire(CLI, command=sys.argv[1:2])
    else:
        fire.Fire(CLI)


if __name__ == "__main__":
    script()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright Toolkit Authors

"""V1DBHandler."""


from copy import deepcopy
import hashlib
import logging

import numpy as np
from tqdm import tqdm
import vaex

from pydtk.utils.utils import load_config
from pydtk.utils.utils import dicts_to_listed_dict_2d
from pydtk.utils.utils import listed_dict_to_dict_1d
from pydtk.utils.utils import dtype_string_to_dtype_object
from pydtk.utils.utils import serialize_dict_1d
from pydtk.utils.utils import deserialize_dict_1d


class BaseDBHandler(object):
    """Handler for db."""

    __version__ = 'v1'
    df_name = 'base_df'

    def __init__(self, path_to_db_files, path_to_output_db_file=None):
        """Initialize V1BaseDBHandler.

        Args:
            path_to_db_files (list): list fo db-files with extension '.arrow'
            path_to_output_db_file (str): output database for saving

        """
        if isinstance(path_to_db_files, str):
            path_to_db_files = [path_to_db_files]

        # Check file extension
        for file in path_to_db_files:
            if not file.endswith('.arrow'):
                raise ValueError('DB file path must ends with ".arrow"')

        self.path_to_db_file = path_to_db_files[0]
        if path_to_output_db_file is not None:
            self.path_to_db_file = path_to_output_db_file
        self._cursor = 0
        self.logger = logging.getLogger(__name__)

        # load config
        self._config = load_config(self.__version__)
        self.columns = self._config[self.df_name]['columns']

        # initialize dataframe
        try:
            self._df = vaex.open_many(path_to_db_files)
        except IOError:
            self._df = None

    def __iter__(self):
        """Return iterator."""
        return self

    def __next__(self):
        """Return the next item."""
        if self._df is None:
            raise EOFError
        if self._cursor >= len(self.df):
            raise EOFError

        # Grab data
        data = self.df.take([self._cursor]).to_dict()

        # Delete internal column
        if 'uuid_in_df' in data.keys():
            del data['uuid_in_df']

        # Post-processes
        data = listed_dict_to_dict_1d(data)
        data = deserialize_dict_1d(data)
        for key, value in data.items():
            if isinstance(value, np.ndarray):
                if np.ma.is_masked(value):
                    data.update({key: None})

        # Increment
        self._cursor += 1

        return data

    def _get_uuid_from_item(self, item):
        """Return UUID of the given item.

        Args:
            item (dict): dict containing data

        Returns:
            (str): UUID

        """
        pre_hash = ''.join([
            '{:.09f}'.format(item[c['name']]) if isinstance(item[c['name']], float) else str(item[c['name']])
            for c in self.columns
            if c['name'] in item.keys()
        ])
        pre_hash = pre_hash.encode('utf-8')
        uuid = hashlib.md5(pre_hash).hexdigest()
        return uuid

    def _preprocess_list_of_dicts(self, data_in):
        """Preprocess list of dicts.

        Args:
            data_in (list): list of dicts containing data

        Returns:
            (dict): dict of lists = listed dict

        """
        data = deepcopy(data_in)

        # Serialize (convert list to str)
        self.logger.info('(Preprocess) Serializing...')
        for item in tqdm(data):
            item = serialize_dict_1d(item)

            # Add df_uuid
            item['uuid_in_df'] = self._get_uuid_from_item(item)

            # Add missing columns
            for column in self.columns:
                column_name, column_dtype = column['name'], column['dtype']
                if column_name not in item.keys():
                    item.update({column_name: None})
                else:
                    dtype_obj = dtype_string_to_dtype_object(column_dtype)
                    if item[column_name] is not None and not isinstance(item[column_name], dtype_obj):
                        try:
                            item[column_name] = dtype_obj(item[column_name])
                        except ValueError:
                            item[column_name] = np.nan

        # Convert dict to listed dict
        self.logger.info('(Preprocess) Converting...')
        data = dicts_to_listed_dict_2d(data)

        return data

    def _append_listed_dict_to_df(self, data, check_unique=True):
        """Append pre-processed dict to self._df.

        Args:
            data (dict): data to add
            check_unique (bool): if True, it will be checked that the data is unique in the db

        """
        if self._df is None:
            self._df = vaex.from_dict(data)
        else:
            if check_unique:
                # TODO: support unique-check for multiple items
                df_uuid = data['uuid_in_df'][0]
                if len(self.df[self.df.uuid_in_df.str.equals(df_uuid)]) > 0:
                    logging.warning('Given data already exist in dataframe: {}'.format(df_uuid))
                    return
            self.df = self.df.concat(vaex.from_dict(data))

    def add_data(self, data_in, **kwargs):
        """Add data to db.

        Args:
            data_in (dict): a dict containing data

        """
        self.add_list_of_data([data_in], **kwargs)

    def add_list_of_data(self, data_in, **kwargs):
        """Add list of data to db.

        Args:
            data_in (list): a list of dicts containing data

        """
        data = self._preprocess_list_of_dicts(data_in)
        self.logger.info('Adding data to DB...')
        self._append_listed_dict_to_df(data, **kwargs)
        self.logger.info('Successfully finished adding data to DB')

    def save(self, path=None):
        """Save data to storage.

        Args:
            path (str): database path to save (default: overwrite to the original file)

        """
        if path is None:
            path = self.path_to_db_file
        self._df.export(path)

    @property
    def df(self):
        """Return df."""
        if self._df is None:
            return vaex.dataframe.DataFrameLocal('df',
                                                 self.path_to_db_file,
                                                 [c['name'] for c in self.columns] + ['uuid_in_df'])
        return self._df

    @df.setter
    def df(self, value):
        """Setter for self.df."""
        self._df = value



"""Search engines."""

from abc import ABCMeta
from typing import TypeVar

from .. import BaseDBHandler, MetaDBHandler, TimeSeriesDBHandler

T = TypeVar(
    'T',
    BaseDBHandler,
    MetaDBHandler,
    TimeSeriesDBHandler
)


class BaseDBSearchEngine(metaclass=ABCMeta):
    """Base DB Search Engine class."""

    def __init__(self, db_handler: T):
        """Initialize BaseDBSearchEngine class.

        Args:
            db_handler (T): database handler

        """
        self._db_handler = db_handler
        self._conditions: [str] = []

    def search(self):
        """Execute SQL query in database handler.

        Returns:
            (pandas.DataFrame): search result

        """
        self._db_handler.read(query=self.query)
        return self._db_handler.df

    def clear(self):
        """Clean search conditions."""
        self._conditions = []

    @property
    def condition(self):
        """Return search condition.

        Returns:
            (str): search condition as string

        """
        return ' and '.join(self._conditions)

    @property
    def select(self):
        """Return select query.

        Returns:
            (str): query

        """
        return 'select * from {}'.format(self._db_handler.df_name)

    @property
    def query(self):
        """Return query.

        Returns:
            (str): query

        """
        if self.condition == '':
            return self.select
        return '{0} where {1}'.format(self.select, self.condition)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright Toolkit Authors

from abc import ABC

from pydtk.models import BaseModel, register_model
import numpy as np
import pandas as pd
import sys
import datetime


@register_model(priority=1)
class GenericCsvModel(BaseModel, ABC):
    """A generic model for a csv file."""

    _content_type = None  # allow any content-type
    _data_type = None   # allow any data-type
    _file_extensions = ['.csv']
    _contents = None

    def __init__(self, **kwargs):
        super(GenericCsvModel, self).__init__(**kwargs)

    def _load(self, path, start_timestamp=None, end_timestamp=None, **kwargs):
        """Load a csv file.

        Args:
            path (str): path to a csv file
            start_timestamp (float): timestamp to start loading (not supported)
            end_timestamp (float): timestamp to end loading (not supported)

        """
        if start_timestamp is not None and end_timestamp is not None:
            raise ValueError('Specifying time-range to load is not supported in GenericCsvModel')
        data = pd.read_csv(path, header=None).to_numpy()
        self.data = data

    def _save(self, path, **kwargs):
        """Save ndarray data to a csv file.

        Args:
            path (str): path to the output csv file

        """
        data = pd.DataFrame(self.data)
        data.to_csv(path, header=False, index=False)

    @property
    def timestamps(self):
        """Return timestamps as ndarray."""
        # this is prototype
        return self.data

    def to_ndarray(self):
        """Return data as ndarray."""
        return self.data

    @classmethod
    def generate_contents_meta(cls, path, content_key='content'):
        """Generate contents metadata.

        Args:
            path (str): File path
            content_key (str): Key of content

        Returns:
            (dict): contents metadata

        """
        # Load file
        data = pd.read_csv(path)
        columns = data.columns.tolist()

        # Generate metadata
        contents = {content_key: {'columns': columns, 'tags': ['csv']}}

        return contents

    @classmethod
    def generate_timestamp_meta(cls, path):
        """Generate contents metadata.

        Args:
            path (str): File path

        Returns:
            (list): [start_timestamp, end_timestamp]

        """
        raise NotImplementedError


@register_model(priority=2)
class CameraTimestampCsvModel(GenericCsvModel, ABC):
    """A model for a csv file containing camera timestamps."""

    _contents = {'camera/.*': {'tags': ['.*']}}
    _columns = ['timestamp']

    def __init__(self, **kwargs):
        super(GenericCsvModel, self).__init__(**kwargs)

    def _load(self, path, start_timestamp=None, end_timestamp=None, **kwargs):
        """Load a csv file.

        Args:
            path (str): path to a csv file
            start_timestamp (float): timestamp to start loading (not supported)
            end_timestamp (float): timestamp to end loading (not supported)

        """
        if start_timestamp is None:
            start_timestamp = self.metadata.data['start_timestamp']
        if end_timestamp is None:
            end_timestamp = self.metadata.data['end_timestamp']

        # load csv
        super()._load(path=path, **kwargs)

        # filter
        start_msec, end_msec = start_timestamp * 1000, end_timestamp * 1000  # sec. -> msec.
        data = self.data
        data = data[np.logical_and(data[:, 0] >= start_msec, data[:, 0] <= end_msec), 0]

        # Convert unit (msec. -> sec.)
        # Note: CSV file timestamps in "Driving behavior DB" is recorded in msec.
        data = data.astype(np.float) * pow(10, -3)

        self.data = data

    def to_ndarray(self):
        """Return data as ndarray."""
        return self.data

    @property
    def timestamps(self):
        """Return timestamps as ndarray."""
        return self.data


@register_model(priority=3)
class AnnotationCsvModel(GenericCsvModel, ABC):
    """A model for a csv file containing annotations."""

    _contents = {'.*annotation': {'tags': ['.*']}}
    _data_type = "annotation"
    _columns = ['Record_ID', 'Annotator_ID', 'Risk_score', 'Subjective_risk_score',
                'Scene_description', 'Risk_factor', 'Environmental_tag', 'Behavior_tag']
    _nan_convert_map = {'Risk_factor': ''}

    def __init__(self, **kwargs):
        super(GenericCsvModel, self).__init__(**kwargs)

    def _load(self, path, start_timestamp=None, end_timestamp=None, **kwargs):
        """Load a csv file.

        Args:
            path (str): path to a csv file

        """
        data = pd.read_csv(path)
        self.data = data

    def _save(self, path, **kwargs):
        """Save ndarray data to a csv file.

        Args:
            path (str): path to the output csv file

        """
        data = pd.DataFrame(self.data)
        data.to_csv(path, header=True, index=False)

    def to_ndarray(self):
        """Return data as ndarray."""
        return self.data.fillna(self._nan_convert_map).to_numpy()

    @property
    def timestamps(self):
        """Return timestamps as ndarray."""
        return np.ndarray([])


@register_model(priority=3)
class ForecastCsvModel(GenericCsvModel, ABC):
    """A model for a csv file containing annotations."""

    _contents = {".*forecast": {"tags": [".*"]}}
    _data_type = "forecast"
    _columns = ["DATE", "TIME", "q=0.01", "q=0.25", "q=0.50", "q=0.75", "q=0.99"]

    def __init__(self, **kwargs):
        super(GenericCsvModel, self).__init__(**kwargs)

    def _load(self, path, start_timestamp=None, end_timestamp=None, **kwargs):
        """Load a csv file.

        Args:
            path (str): path to a csv file

        """
        data = pd.read_csv(path)

        if start_timestamp is None and end_timestamp is None:
            self.data = data
            return

        start = start_timestamp if start_timestamp else 0.0
        end = end_timestamp if end_timestamp else sys.float_info.max

        tmp = []
        for _, sample in data.iterrows():
            current = datetime.datetime.strptime(
                sample["DATE"] + " " + sample["TIME"], "%Y/%m/%d %H:%M"
            ).timestamp()
            if start <= current and current <= end:
                tmp.append(sample)

        self.data = pd.DataFrame(tmp)

    def _save(self, path, **kwargs):
        """Save ndarray data to a csv file.

        Args:
            path (str): path to the output csv file

        """
        data = pd.DataFrame(self.data)
        data.to_csv(path, header=True, index=False)

    def to_ndarray(self):
        """Return data as ndarray."""
        return self.data.to_numpy()

    @property
    def timestamps(self):
        """Return timestamps as ndarray."""
        return np.ndarray([])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright Toolkit Authors

"""V4DBHandler."""

import logging
import importlib
import os

DB_ENGINES = {}  # key: db_class, value: dict( key: db_engine, value: handler )

logger = logging.getLogger(__name__)


def register_engines():
    """Register engines."""
    for filename in os.listdir(os.path.join(os.path.dirname(__file__))):
        if not os.path.isfile(os.path.join(os.path.dirname(__file__), filename)):
            continue
        if filename == '__init__.py':
            continue

        try:
            engine_name = str(os.path.splitext(filename)[0])
            module_name = os.path.join('pydtk.db.v4.engines', engine_name).replace(os.sep, '.')
            DB_ENGINES[engine_name] = importlib.import_module(module_name)
        except ModuleNotFoundError:
            logger.warning('Failed to load DB-engine {}'.format(filename))


register_engines()



"""Search engines."""

from abc import ABCMeta
import importlib
import logging
import os
from typing import TypeVar

from .. import DBHandler
from .. import MetaDBHandler
from .. import TimeSeriesDBHandler
from .. import TimeSeriesCassandraDBHandler
from .. import StatisticsDBHandler
from .. import StatisticsCassandraDBHandler

T = TypeVar(
    'T',
    DBHandler,
    MetaDBHandler,
    TimeSeriesDBHandler,
    TimeSeriesCassandraDBHandler,
    StatisticsDBHandler,
    StatisticsCassandraDBHandler
)

DB_SEARCH_ENGINES = {}    # key: db_handler class, value: search engine


def register_engines():
    """Register engines."""
    for filename in os.listdir(os.path.join(os.path.dirname(__file__))):
        if not os.path.isfile(os.path.join(os.path.dirname(__file__), filename)):
            continue
        if filename == '__init__.py':
            continue

        try:
            importlib.import_module(
                os.path.join('pydtk.db.v3.search_engines',
                             os.path.splitext(filename)[0]).replace(os.sep, '.')
            )
        except ModuleNotFoundError:
            logging.debug('Failed to load handlers in {}'.format(filename))


def register_engine(db_handlers: [T]):
    """Register a DB-search-engine.

    Args:
        db_handlers ([T]): database handler class

    """
    def decorator(cls):
        for db_handler in db_handlers:
            if db_handler not in DB_SEARCH_ENGINES.keys():
                DB_SEARCH_ENGINES.update({db_handler: cls})
        return cls
    return decorator


@register_engine(db_handlers=[
    MetaDBHandler,
    TimeSeriesDBHandler,
    StatisticsDBHandler
])
class BaseDBSearchEngine(metaclass=ABCMeta):
    """Base DB Search Engine class."""

    def __new__(cls, db_handler, **kwargs) -> object:
        """Create object.

        Args:
            db_handler (T): database handler
            **kwargs: DB-handler specific arguments

        Returns:
            (object): the corresponding handler object

        """
        if cls is BaseDBSearchEngine:
            engine = cls._get_search_engine(db_handler)
            return super(BaseDBSearchEngine, cls).__new__(engine)
        else:
            return super(BaseDBSearchEngine, cls).__new__(cls)

    @classmethod
    def _get_search_engine(cls, db_handler):
        """Returns an appropriate search engine.

        Args:
            db_handler (T): database handler

        Returns:
            (handler): database handler object

        """
        for _db_handler, engine in DB_SEARCH_ENGINES.items():
            if type(db_handler) is _db_handler:
                return engine

        raise ValueError('Unsupported handler: {}'.format(db_handler))

    def __init__(self, db_handler: T):
        """Initialize BaseDBSearchEngine class.

        Args:
            db_handler (T): database handler

        """
        self._db_handler = db_handler
        self._conditions: [str] = []

    def add_condition(self, condition: str):
        """Add a search condition.

        Args:
            condition (str): search condition
                             (e.g. '"/vehicle/acceleration/accel_linear_x/mean" > 0')

        """
        if '"' not in condition:
            raise ValueError('column name must be quoted '
                             '(e.g. "/vehicle/acceleration/accel_linear_x/mean" > 0)')
        self._conditions.append(condition)

    def search(self):
        """Execute SQL query in database handler.

        Returns:
            (pandas.DataFrame): search result

        """
        self._db_handler.read(query=self.query)
        return self._db_handler.df

    def clear(self):
        """Clean search conditions."""
        self._conditions = []

    @property
    def condition(self):
        """Return search condition.

        Returns:
            (str): search condition as string

        """
        return ' and '.join(self._conditions)

    @property
    def select(self):
        """Return select query.

        Returns:
            (str): query

        """
        return 'select * from "{}"'.format(self._db_handler.df_name)

    @property
    def query(self):
        """Return query.

        Returns:
            (str): query

        """
        if self.condition == '':
            return self.select
        return '{0} where {1}'.format(self.select, self.condition)


register_engines()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright Toolkit Authors

from abc import ABC
import sys

from pydtk.models import BaseModel, register_model
import numpy as np

from pandas import DataFrame
from pyntcloud import PyntCloud
from pypcd import pypcd


@register_model(priority=1)
class PCDModel(BaseModel, ABC):
    """A generic model for a .pcd file."""

    _content_type = None  # allow any content-type
    _data_type = None   # allow any data-type
    _file_extensions = ['.pcd']
    _contents = '.*'
    _columns = ['x', 'y', 'z']
    data: DataFrame

    def __init__(self, **kwargs):
        super(PCDModel, self).__init__(**kwargs)

    def _load(self, path, start_timestamp=None, end_timestamp=None, **kwargs):
        """Load a csv file.

        Args:
            path (str): path to a csv file
            start_timestamp (float): timestamp to start loading (not supported)
            end_timestamp (float): timestamp to end loading (not supported)

        """
        if start_timestamp is not None and end_timestamp is not None:
            sys.stderr.write('Warning: Specifying time-range to load is not supported in PCDModel\n')
        try:
            cloud = PyntCloud.from_file(path)
            data = cloud.points
            columns = data.columns.tolist()
        except NotImplementedError:
            cloud = pypcd.PointCloud.from_path(path)
            data = DataFrame(cloud.pc_data)
            columns = cloud.fields
        self.data = data
        self._columns = columns

    def _save(self, path, compression='binary', **kwargs):
        """Save ndarray data to a csv file.

        Args:
            path (str): path to the output csv file

        """
        # TODO: Use pyntcloud instead of pypcd when saving .pcd files is supported

        assert type(self.data) == DataFrame
        dtypes = [dtype.name for dtype in self.data.dtypes.tolist()]
        points = np.core.records.fromarrays(self.to_ndarray().transpose(),
                                            names=', '.join(self._columns),
                                            formats=', '.join(dtypes))
        cloud = pypcd.PointCloud.from_array(points)
        cloud.save_pcd(path, compression=compression)

    @property
    def timestamps(self):
        """Return timestamps as ndarray."""
        return np.array([0])

    def to_ndarray(self):
        """Return data as ndarray."""
        return self.data.to_numpy()

    def from_ndarray(self, data, columns=None):
        """Set data from ndarray.

        Args:
            data (np.array): data

        """
        if columns is None:
            columns = self._columns
        assert len(data.shape) == 2
        assert len(columns) == data.shape[-1]

        self.data = DataFrame(data, columns=columns)
        self._columns = columns

    @classmethod
    def generate_contents_meta(cls, path, content_key='content'):
        """Generate contents metadata.

        Args:
            path (str): File path
            content_key (str): Key of content

        Returns:
            (dict): contents metadata

        """
        raise NotImplementedError

    @classmethod
    def generate_timestamp_meta(cls, path):
        """Generate contents metadata.

        Args:
            path (str): File path

        Returns:
            (list): [start_timestamp, end_timestamp]

        """
        raise NotImplementedError



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright Toolkit Authors

"""pydtk modules."""

from abc import ABCMeta

from pydtk.models import MODELS_BY_PRIORITY, MetaDataModel
from pydtk.preprocesses import PassThrough
from pydtk.io.errors import NoModelMatchedError


class BaseFileWriter(metaclass=ABCMeta):
    """Base file writer."""

    _model = None   # model used for loading a file

    def __init__(self, **kwargs):
        self.preprocesses = [PassThrough()]

    @classmethod
    def _select_model(cls, file_metadata):
        """Select a proper model based on the given file-metadata.

        Args:
            file_metadata (object): an MetaDataModel object

        """
        priorities = MODELS_BY_PRIORITY.keys()
        for priority in sorted(priorities, reverse=True):
            for model in MODELS_BY_PRIORITY[priority]:
                if model.is_loadable(**file_metadata.data):
                    return model
        raise NoModelMatchedError('No suitable model found for loading data: {}'.
                                  format(file_metadata))

    @property
    def model(self):
        """Return model."""
        return self._model

    @model.setter
    def model(self, model):
        self._model = model

    def write(self,
              metadata=None,
              data=None,
              model_kwargs=None,
              **kwargs
              ):
        """Write a file which corresponds to the given metadata.

        Args:
            metadata (dict or MetaDataModel): metadata of the data to save
            data (numpy array): data
            model_kwargs (dict): kwargs to pass to the selected model

        Returns:
            void

        """
        if model_kwargs is None:
            model_kwargs = {}

        # Check metadata is valid
        if metadata is None:
            raise ValueError('Metadata must be specified')
        if type(metadata) is dict:
            if 'path' not in metadata.keys():
                raise ValueError('Metadata must have path key')
            else:
                metadata = MetaDataModel(metadata)
        elif type(metadata) is MetaDataModel:
            pass
        else:
            raise ValueError('Type of metadata must be dict or MetaDataModel')

        metadata.save(metadata.data['path'] + metadata._file_extensions[0])

        # Select a suitable model and save data
        self.model = self._select_model(metadata)
        self.model = self.model(metadata=metadata, data=data, **model_kwargs)
        self.model.save()



import sys

if sys.version_info[0] < 3:
    raise Exception('Python >= 3.7 is required')

if sys.version_info[1] < 7:
    raise Exception('Python >= 3.7 is required')

if sys.version_info[1] == 7:
    from .pql_python37 import pql
else:
    from .pql_python38 import pql



'''
TODO:

optimize adds, multiplies, 'or' and 'and' as they can accept more than two values
validate type info on specific functions
'''
from .matching import AstHandler, ParseError, DateTimeFunc

class AggregationParser(AstHandler):

    FUNC_TO_ARGS = {'concat': '+', # more than 1
                    'strcasecmp': 2,
                    'substr': 3,
                    'toLower': 1,
                    'toUpper': 1,

                    'dayOfYear': 1,
                    'dayOfMonth': 1,
                    'dayOfWeek': 1,
                    'year': 1,
                    'month': 1,
                    'week': 1,
                    'hour': 1,
                    'minute': 1,
                    'second': 1,
                    'millisecond': 1,
                    
                    'date': 1,

                    'cmp': 2,

                    'ifnull': 2}

    SPECIAL_VALUES = {'False': False,
                      'false': False,
                      'True': True,
                      'true': True,
                      'None': None,
                      'null': None}
    
    def handle_Str(self, node):
        return node.s

    def handle_Num(self, node):
        return node.n

    def handle_Name(self, node):
        return self.SPECIAL_VALUES.get(node.id, '$' + node.id)

    def handle_NameConstant(self,node):
        return self.SPECIAL_VALUES.get(str(node.value),node.value)

    def handle_Attribute(self, node):
        return '${0}.{1}'.format(self.handle(node.value), node.attr).replace('$$', '$')

    def handle_UnaryOp(self, op):
        return {self.handle(op.op): self.handle(op.operand)}

    def handle_IfExp(self, op):
        return {'$cond': [self.handle(op.test),
                          self.handle(op.body),
                          self.handle(op.orelse)]}

    def handle_Call(self, node):
        name = node.func.id
        if name == 'date':
            return DateTimeFunc().handle_date(node)
        if name not in self.FUNC_TO_ARGS:
            raise ParseError('Unsupported function ({0}).'.format(name),
                             col_offset=node.col_offset)
        if len(node.args) != self.FUNC_TO_ARGS[name] and \
           self.FUNC_TO_ARGS[name] != '+' or len(node.args) == 0:
            raise ParseError('Invalid number of arguments to function {0}'.format(name),
                             col_offset=node.col_offset)

        # because of SERVER-9289 the following fails: {'$year': {'$add' :['$time_stamp', 1]}}
        # wrapping both single arg functions in a list solves it: {'$year': [{'$add' :['$time_stamp', 1]}]}
        return {'$' + node.func.id: list(map(self.handle, node.args))}

    def handle_BinOp(self, node):
        return {self.handle(node.op): [self.handle(node.left),
                                       self.handle(node.right)]}

    def handle_Not(self, not_node):
        return '$not'

    def handle_And(self, op):
        return '$and'

    def handle_Or(self, op):
        return '$or'

    def handle_BoolOp(self, op):
        return {self.handle(op.op): list(map(self.handle, op.values))}

    def handle_Compare(self, node):
        if len(node.ops) != 1:
            raise ParseError('Invalid number of comparators: {0}'.format(len(node.ops)),
                             col_offset=node.comparators[1].col_offset)
        return {self.handle(node.ops[0]): [self.handle(node.left),
                                           self.handle(node.comparators[0])]}

    def handle_Gt(self, node):
        return '$gt'
        
    def handle_Lt(self,node):
        return '$lt'
        
    def handle_GtE(self, node):
        return '$gte'
        
    def handle_LtE(self, node):
        return '$lte'

    def handle_Eq(self, node):
        return '$eq'
        
    def handle_NotEq(self, node):
        return '$ne'

    def handle_Add(self, node):
        return '$add'

    def handle_Sub(self, node):
        return '$subtract'

    def handle_Mod(self, node):
        return '$mod'

    def handle_Mult(self, node):
        return '$multiply'

    def handle_Div(self, node):
        return '$divide'

class AggregationGroupParser(AstHandler):
    GROUP_FUNCTIONS = ['addToSet', 'push', 'first', 'last',
                       'max', 'min', 'avg', 'sum']
    def handle_Call(self, node):
        if len(node.args) != 1:
            raise ParseError('The {0} group aggregation function accepts one argument'.format(node.func.id),
                             col_offset=node.col_offset)
        if node.func.id not in self.GROUP_FUNCTIONS:
            raise ParseError('Unsupported group function: {0}'.format(node.func.id),
                             col_offset=node.col_offset,
                             options=self.GROUP_FUNCTIONS)
        return {'$' + node.func.id: AggregationParser().handle(node.args[0])}




#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright Toolkit Authors

"""Base Statistic Calculation module."""

import numpy as np
from abc import ABCMeta


class UnsupportedOperationError(BaseException):
    """Error for unsupported file."""

    pass


class BaseCalculator(metaclass=ABCMeta):
    """Base Calculator."""

    def __init__(self, target_span=60.0, sync_timestamps=False):
        """Initialize Base Statistics Calculator.

        Args:
            target_span (float): interval of statistics calculation
            sync_timestamps (bool): if True, the output timestamps will
                                          start from 'timestamp // span * span'

        """
        self.target_span = target_span
        self.sync_timestamps = sync_timestamps
        self.operations = None

    def divide(self, timestamps, data):
        """Divide data with target span.

        Args:
            timestamps (ndarray): timestamps [sec]
            data (ndarray): signal to downsample

        Returns:
            divided_timestamps (ndarray): timestamps [sec]
            divided_data (list): divided data list

        """
        # Initialize the output array and buffers
        divided_timestamps, divided_data = [], []
        buffer_timestamps, buffer_data = [], []

        # Calculate the first timestamp-index
        fps_previous_index = timestamps[0] // float(self.target_span) \
            if len(timestamps) > 0 else 0

        # Read over the data to store timestamps and data
        for timestamp, value in zip(timestamps, data):
            fps_current_index = timestamp // float(self.target_span)

            if fps_current_index == fps_previous_index:
                # Store data into the buffer in case of the same timestamp-index
                buffer_timestamps.append(timestamp)
                buffer_data.append(value)
            else:
                # Move data in the buffer to the output array
                if self.sync_timestamps:
                    divided_timestamps.append(fps_previous_index * self.target_span)
                else:
                    divided_timestamps.append(buffer_timestamps[0])
                divided_data.append(np.array(buffer_data))

                # Store the new data having a different timestamp-index into the buffer
                buffer_timestamps, buffer_data = [timestamp], [value]
                fps_previous_index = fps_current_index

        # Handle the data remaining in the buffer
        if len(buffer_data) > 0:
            if self.sync_timestamps:
                divided_timestamps.append(
                    buffer_timestamps[0] // self.target_span * self.target_span
                )
            else:
                divided_timestamps.append(buffer_timestamps[0])
            divided_data.append(np.array(buffer_data))

        divided_timestamps = np.array(divided_timestamps)
        return divided_timestamps, divided_data

    def mean(self, timestamps, data):
        """Divide and return means of divided data."""
        raise UnsupportedOperationError("Model '{0}' does not support operation: {1}".
                                        format(type(self).__name__, "mean"))

    def max(self, timestamps, data):
        """Divide and return maximum of divided data."""
        raise UnsupportedOperationError("Model '{0}' does not support operation: {1}".
                                        format(type(self).__name__, "max"))

    def min(self, timestamps, data):
        """Divide and return minimum of divided data."""
        raise UnsupportedOperationError("Model '{0}' does not support operation: {1}".
                                        format(type(self).__name__, "min"))

    def count(self, timestamps, data):
        """Divide and return count of True in divided data."""
        raise UnsupportedOperationError("Model '{0}' does not support operation: {1}".
                                        format(type(self).__name__, "count"))


class FloatCalculator(BaseCalculator):
    """Calculator for data of float."""

    def __init__(self, target_span=60.0, **kwargs):
        super().__init__(target_span, **kwargs)
        self.operations = ["mean", "max", "min"]

    def mean(self, timestamps, data):
        """Calculate mean.

        Args:
            timestamps (ndarray): timestamps [sec]
            data (ndarray): input data

        Returns:
            index_timestamps (ndarray): timestamps [sec]
            stat_data (ndarray): mean of input data

        """
        index_timestamps, divided_data = self.divide(timestamps, data)
        data_dim = data.ndim
        stat_data = [values.mean(axis=0) if values.ndim == data_dim else values
                     for values in divided_data]
        stat_data = np.array(stat_data)
        return index_timestamps, stat_data

    def max(self, timestamps, data):
        """Calculate max.

        Args:
            timestamps (ndarray): timestamps [sec]
            data (ndarray): input data

        Returns:
            index_timestamps (ndarray): timestamps [sec]
            stat_data (ndarray): max of input data

        """
        index_timestamps, divided_data = self.divide(timestamps, data)
        data_dim = data.ndim
        stat_data = [values.max(axis=0) if values.ndim == data_dim else values
                     for values in divided_data]
        stat_data = np.array(stat_data)
        return index_timestamps, stat_data

    def min(self, timestamps, data):
        """Calculate minimum.

        Args:
            timestamps (ndarray): timestamps [sec]
            data (ndarray): input data

        Returns:
            index_timestamps (ndarray): timestamps [sec]
            stat_data (ndarray): min of input data

        """
        index_timestamps, divided_data = self.divide(timestamps, data)
        data_dim = data.ndim
        stat_data = [values.min(axis=0) if values.ndim == data_dim else values
                     for values in divided_data]
        stat_data = np.array(stat_data)
        return index_timestamps, stat_data


class BoolCalculator(BaseCalculator):
    """Calculator for data of float."""

    def __init__(self, target_span=60.0, **kwargs):
        super().__init__(target_span, **kwargs)
        self.operations = ["mean", "max", "min"]

    def count(self, timestamps, data):
        """Count True.

        Args:
            timestamps (ndarray): timestamps [sec]
            data (ndarray): input data

        Returns:
            index_timestamps (ndarray): timestamps [sec]
            stat_data (ndarray): count of true data

        """
        index_timestamps, divided_data = self.divide(timestamps, data)
        data_dim = data.ndim
        stat_data = [values.sum(axis=0) if values.ndim == data_dim else values
                     for values in divided_data]
        stat_data = np.array(stat_data)
        return index_timestamps, stat_data

    def mean(self, timestamps, data):
        """Average of true counts during the span.

        Args:
            timestamps (ndarray): timestamps [sec]
            data (ndarray): input data

        Returns:
            index_timestamps (ndarray): timestamps [sec]
            stat_data (ndarray): count of true data

        """
        return self.count(timestamps, data)

    def max(self, timestamps, data):
        """Calculate max.

        Args:
            timestamps (ndarray): timestamps [sec]
            data (ndarray): input data

        Returns:
            index_timestamps (ndarray): timestamps [sec]
            stat_data (ndarray): max of input data

        """
        index_timestamps, divided_data = self.divide(timestamps, data)
        data_dim = data.ndim
        stat_data = [values.max(axis=0) if values.ndim == data_dim else values
                     for values in divided_data]
        stat_data = np.array(stat_data)
        return index_timestamps, stat_data

    def min(self, timestamps, data):
        """Calculate minimum.

        Args:
            timestamps (ndarray): timestamps [sec]
            data (ndarray): input data

        Returns:
            index_timestamps (ndarray): timestamps [sec]
            stat_data (ndarray): min of input data

        """
        index_timestamps, divided_data = self.divide(timestamps, data)
        data_dim = data.ndim
        stat_data = [values.min(axis=0) if values.ndim == data_dim else values
                     for values in divided_data]
        stat_data = np.array(stat_data)
        return index_timestamps, stat_data



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright Toolkit Authors

from abc import ABC

import cv2
import numpy as np
import pyzstd
import ros_numpy
import sensor_msgs.msg

from pydtk.models import register_model
from pydtk.models.rosbag import GenericRosbagModel as _GenericRosbagModel


@register_model(priority=1)
class GenericZstdRosbagModel(_GenericRosbagModel, ABC):
    """A generic model for a zstandard rosbag file."""

    _file_extensions = ['.zst']

    def _load(self, path, **kwargs):
        """Load a zstandard rosbag file.

        Args:
            path (str): path to a rosbag file

        """
        with pyzstd.open(path, 'rb') as f:
            f.mode = 'rb'
            super()._load(path=f, **kwargs)

    def _load_as_generator(self, path, **kwargs):
        """Load a zstandard rosbag file for each sample.

        Args:
            path (str): path to a rosbag file

        """
        with pyzstd.open(path, 'rb') as f:
            f.mode = 'rb'
            generator = super()._load_as_generator(path=f, **kwargs)
            for sample in generator:
                yield sample

    def _save(self, path, contents=None, **kwargs):
        """Save ndarray data to a zstandard rosbag file.

        Args:
            path (str): path to the output zst file
            contents (str or dict): topic name

        """
        # TODO: implementation
        pass

    @classmethod
    def generate_contents_meta(cls, path, **kwargs):
        """Generate contents metadata.

        Args:
            path (str): File path

        Returns:
            (dict): contents metadata

        """
        with pyzstd.open(path, 'rb') as f:
            f.mode = 'rb'
            return super().generate_contents_meta(path=f, **kwargs)

    @classmethod
    def generate_timestamp_meta(cls, path):
        """Generate contents metadata.

        Args:
            path (str): File path

        Returns:
            (list): [start_timestamp, end_timestamp]

        """
        with pyzstd.open(path, 'rb') as f:
            f.mode = 'rb'
            return super().generate_timestamp_meta(path=f)


@register_model(priority=2)
class SensorMsgsCompressedImageZstdRosbagModel(GenericZstdRosbagModel, ABC):
    """A model for a rosbag file containing sensor_msgs/Range."""

    _contents = {'.*': {'msg_type': 'sensor_msgs/CompressedImage'}}
    _columns = ['red', 'green', 'blue']

    @staticmethod
    def msg_to_data(msg, resize_rate=1.0, **kwargs):
        """Convert a message to data."""
        jpg = np.fromstring(msg.data, np.uint8)
        image = cv2.imdecode(jpg, cv2.IMREAD_COLOR)
        if resize_rate != 1.0:
            image = cv2.resize(image, dsize=None, fx=resize_rate,
                               fy=resize_rate, interpolation=cv2.INTER_LINEAR)
        image = image[:, :, ::-1]  # Convert BGR to RGB
        image = image.transpose((2, 0, 1))  # Reshape: [H, W, C] -> [C, H, W]
        return image

    def to_ndarray(self):
        """Return data as ndarray."""
        return np.array(self.data['data'])


@register_model(priority=2)
class SensorMsgsPointCloud2ZstdRosbagModel(GenericZstdRosbagModel, ABC):
    """A model for a rosbag file containing sensor_msgs/PointCloud2."""

    _contents = {'.*': {'msg_type': 'sensor_msgs/PointCloud2'}}
    _config = {'fields': ('x', 'y', 'z')}

    def __init__(self, fields=('x', 'y', 'z'), **kwargs):
        super(SensorMsgsPointCloud2ZstdRosbagModel, self).__init__(**kwargs)
        self._config['fields'] = fields

    def msg_to_data(self, msg, **kwargs):
        """Convert a message to data."""
        if msg.__class__.__name__ == "_sensor_msgs__PointCloud2":
            msg.__class__ = sensor_msgs.msg._PointCloud2.PointCloud2
        points = ros_numpy.numpify(msg)[list(self._config['fields'])]
        pointcloud = np.array(points.tolist())
        if 'intensity' in self._config['fields']:
            pointcloud[:, self._config['fields'].index('intensity')] /= 255.0  # scale to [0, 1]
        return pointcloud

    def to_ndarray(self):
        """Return data as ndarray."""
        return np.array(self.data['data'], dtype="object")

    @property
    def columns(self):
        """Return columns."""
        return list(self._config['fields'])
